from __future__ import absolute_import, division, print_function

from collections import Iterator, Mapping
import decimal
import datetime
from functools import reduce, partial
import itertools
import operator
import warnings

import datashape
from datashape import discover, Tuple, Record, DataShape, var, Map
from datashape.predicates import (
    isscalar,
    iscollection,
    isrecord,
    istabular,
    _dimensions,
)
import numpy as np
from odo import resource, odo, append, drop
from odo.utils import ignoring, copydoc
from odo.compatibility import unicode
from pandas import DataFrame, Series, Timestamp


from .expr import Expr, Symbol, ndim
from .expr.expressions import sanitized_dshape
from .dispatch import dispatch
from .compatibility import _strtypes


__all__ = ['Data', 'into', 'to_html', 'data']


names = ('_%d' % i for i in itertools.count(1))
not_an_iterator = []


with ignoring(ImportError):
    import bcolz
    not_an_iterator.append(bcolz.carray)


with ignoring(ImportError):
    import pymongo
    not_an_iterator.append(pymongo.collection.Collection)
    not_an_iterator.append(pymongo.database.Database)


class _Data(Symbol):

    # NOTE: This docstring is meant to correspond to the ``data()`` API, which
    # is why the Parameters section doesn't match the arguments to
    # ``_Data.__init__()``.

    """Bind a data resource to a symbol, for use in expressions and
    computation.

    A ``data`` object presents a consistent view onto a variety of concrete
    data sources.  Like ``symbol`` objects, they are meant to be used in
    expressions.  Because they are tied to concrete data resources, ``data``
    objects can be used with ``compute`` directly, making them convenient for
    interactive exploration.

    Parameters
    ----------
    data_source : object
        Any type with ``discover`` and ``compute`` implementations
    fields : list, optional
        Field or column names, will be inferred from data_source if possible
    dshape : str or DataShape, optional
        DataShape describing input data
    name : str, optional
        A name for the data.

    Examples
    --------
    >>> t = data([(1, 'Alice', 100),
    ...           (2, 'Bob', -200),
    ...           (3, 'Charlie', 300),
    ...           (4, 'Denis', 400),
    ...           (5, 'Edith', -500)],
    ...          fields=['id', 'name', 'balance'])
    >>> t[t.balance < 0].name
        name
    0    Bob
    1  Edith
    """
    _arguments = 'data', 'dshape', '_name'

    def __new__(cls, data, dshape, name=None):
        return super(Symbol, cls).__new__(
            cls,
            data,
            dshape,
            name or (
                next(names)
                if isrecord(dshape.measure) else None
            ),
        )

    def _resources(self):
        return {self: self.data}

    @classmethod
    def _static_identity(cls, data, dshape, _name):
        try:
            # cannot use isinstance(data, Hashable)
            # some classes give a false positive
            hash(data)
        except TypeError:
            data = id(data)
        return cls, data, dshape, _name

    def __repr__(self):
        fmt = "<'{}' data; _name='{}', dshape='{}'>"
        return fmt.format(type(self.data).__name__,
                          self._name,
                          sanitized_dshape(self.dshape))


class InteractiveSymbol(_Data):
    """Deprecated and replaced by the _Data class. ``InteractiveSymbol`` will
    be removed in version 0.10.  ``data`` is the public API for creating
    ``Data`` objects.
    """

    def __new__(cls, *args, **kwargs):
        warnings.warn("InteractiveSymbol has been deprecated in 0.10 and will"
                      " be removed in 0.11.  Use ``data`` to create ``Data``"
                      " objects directly.", DeprecationWarning)
        return data(*args, **kwargs)


def Data(data_source, dshape=None, name=None, fields=None, schema=None, **kwargs):
    warnings.warn("""The `Data` callable has been deprecated in 0.10 and willbe removed in
                   version >= 0.11. It has been renamed `data`.""", DeprecationWarning)
    return data(data_source, dshape=dshape, name=name, fields=fields, schema=schema, **kwargs)


@copydoc(_Data)
def data(data_source, dshape=None, name=None, fields=None, schema=None, **kwargs):
    if schema and dshape:
        raise ValueError("Please specify one of schema= or dshape= keyword"
                         " arguments")

    if isinstance(data_source, _Data):
        return data(data_source.data, dshape, name, fields, schema, **kwargs)

    if schema and not dshape:
        dshape = var * schema
    if dshape and isinstance(dshape, _strtypes):
        dshape = datashape.dshape(dshape)

    if isinstance(data_source, _strtypes):
        data_source = resource(data_source, schema=schema, dshape=dshape, **kwargs)

    if (isinstance(data_source, Iterator) and
            not isinstance(data_source, tuple(not_an_iterator))):
        data_source = tuple(data_source)
    if not dshape:
        dshape = discover(data_source)
        types = None
        if isinstance(dshape.measure, Tuple) and fields:
            types = dshape[1].dshapes
            schema = Record(list(zip(fields, types)))
            dshape = DataShape(*(dshape.shape + (schema,)))
        elif isscalar(dshape.measure) and fields:
            types = (dshape.measure,) * int(dshape[-2])
            schema = Record(list(zip(fields, types)))
            dshape = DataShape(*(dshape.shape[:-1] + (schema,)))
        elif isrecord(dshape.measure) and fields:
            ds = discover(data_source)
            assert isrecord(ds.measure)
            names = ds.measure.names
            if names != fields:
                raise ValueError('data column names %s\n'
                                 '\tnot equal to fields parameter %s,\n'
                                 '\tuse data(data_source).relabel(%s) to rename '
                                 'fields' % (names,
                                             fields,
                                             ', '.join('%s=%r' % (k, v)
                                                       for k, v in
                                                       zip(names, fields))))
            types = dshape.measure.types
            schema = Record(list(zip(fields, types)))
            dshape = DataShape(*(dshape.shape + (schema,)))

    ds = datashape.dshape(dshape)
    return _Data(data_source, ds, name)


@dispatch(_Data, Mapping)
def _subs(o, d):
    return o


@dispatch(Expr)
def compute(expr, **kwargs):
    resources = expr._resources()
    if not resources:
        raise ValueError("No data resources found")
    else:
        return compute(expr, resources, **kwargs)


@dispatch(Expr, _Data)
def compute_down(expr, dta, **kwargs):
    return compute(expr, dta.data, **kwargs)


@dispatch(Expr, _Data)
def pre_compute(expr, dta, **kwargs):
    return pre_compute(expr, dta.data, **kwargs)


def concrete_head(expr, n=10):
    """ Return head of computed expression """
    if not expr._resources():
        raise ValueError("Expression does not contain data resources")
    if not iscollection(expr.dshape):
        return compute(expr)

    head = expr.head(n + 1)

    if not iscollection(expr.dshape):
        return odo(head, object)
    elif isrecord(expr.dshape.measure):
        return odo(head, DataFrame)

    df = odo(head, DataFrame)
    df.columns = [expr._name]
    return df


def _peek_tables(expr, n=10):
    return concrete_head(expr, n).rename(columns={None: ''})


def repr_tables(expr, n=10):
    result = concrete_head(expr, n).rename(columns={None: ''})

    if isinstance(result, (DataFrame, Series)):
        s = repr(result)
        if len(result) > 10:
            s = '\n'.join(s.split('\n')[:-1]) + '\n...'
        return s
    else:
        return result.peek()  # pragma: no cover


def numel(shape):
    if var in shape:
        return None
    if not shape:
        return 1
    return reduce(operator.mul, shape, 1)


def short_dshape(ds, nlines=5):
    s = datashape.coretypes.pprint(ds)
    lines = s.split('\n')
    if len(lines) > 5:
        s = '\n'.join(lines[:nlines]) + '\n  ...'
    return s


def coerce_to(typ, x, odo_kwargs=None):
    try:
        return typ(x)
    except TypeError:
        return odo(x, typ, **(odo_kwargs or {}))


def coerce_scalar(result, dshape, odo_kwargs=None):
    dshape = str(dshape)
    coerce_ = partial(coerce_to, x=result, odo_kwargs=odo_kwargs)
    if 'float' in dshape:
        return coerce_(float)
    if 'decimal' in dshape:
        return coerce_(decimal.Decimal)
    elif 'int' in dshape:
        return coerce_(int)
    elif 'bool' in dshape:
        return coerce_(bool)
    elif 'datetime' in dshape:
        return coerce_(Timestamp)
    elif 'date' in dshape:
        return coerce_(datetime.date)
    elif 'timedelta' in dshape:
        return coerce_(datetime.timedelta)
    else:
        return result


def coerce_core(result, dshape, odo_kwargs=None):
    """Coerce data to a core data type."""
    if iscoretype(result):
        return result
    elif isscalar(dshape):
        result = coerce_scalar(result, dshape, odo_kwargs=odo_kwargs)
    elif istabular(dshape) and isrecord(dshape.measure):
        result = into(DataFrame, result, **(odo_kwargs or {}))
    elif iscollection(dshape):
        dim = _dimensions(dshape)
        if dim == 1:
            result = into(Series, result, **(odo_kwargs or {}))
        elif dim > 1:
            result = into(np.ndarray, result, **(odo_kwargs or {}))
        else:
            msg = "Expr with dshape dimensions < 1 should have been handled earlier: dim={}"
            raise ValueError(msg.format(str(dim)))
    else:
        msg = "Expr does not evaluate to a core return type"
        raise ValueError(msg)

    return result


def _peek(expr):
    # Pure Expressions, not interactive
    if not set(expr._resources().keys()).issuperset(expr._leaves()):
        return expr

    # Scalars
    if ndim(expr) == 0 and isscalar(expr.dshape):
        return coerce_scalar(compute(expr), str(expr.dshape))

    # Tables
    if (ndim(expr) == 1 and (istabular(expr.dshape) or
                             isscalar(expr.dshape.measure) or
                             isinstance(expr.dshape.measure, Map))):
        return _peek_tables(expr, 10)

    # Smallish arrays
    if ndim(expr) >= 2 and numel(expr.shape) and numel(expr.shape) < 1000000:
        return compute(expr)

    # Other
    dat = expr._resources().values()
    if len(dat) == 1:
        dat = list(dat)[0]  # may be dict_values
    return dat


def expr_repr(expr, n=10):
    # Pure Expressions, not interactive
    if not set(expr._resources().keys()).issuperset(expr._leaves()):
        return str(expr)

    # Scalars
    if ndim(expr) == 0 and isscalar(expr.dshape):
        return repr(coerce_scalar(compute(expr), str(expr.dshape)))

    # Tables
    if (ndim(expr) == 1 and (istabular(expr.dshape) or
                             isscalar(expr.dshape.measure) or
                             isinstance(expr.dshape.measure, Map))):
        return repr_tables(expr, 10)

    # Smallish arrays
    if ndim(expr) >= 2 and numel(expr.shape) and numel(expr.shape) < 1000000:
        return repr(compute(expr))

    # Other
    dat = expr._resources().values()
    if len(dat) == 1:
        dat = list(dat)[0]  # may be dict_values

    s = 'Data:       %s' % dat
    if not isinstance(expr, Symbol):
        s += '\nExpr:       %s' % str(expr)
    s += '\nDataShape:  %s' % short_dshape(expr.dshape, nlines=7)

    return s


@dispatch(DataFrame)
def to_html(df):
    return df.to_html()


@dispatch(Expr)
def to_html(expr):
    # Tables
    if not expr._resources() or ndim(expr) != 1:
        return to_html(expr_repr(expr))
    return to_html(concrete_head(expr))


@dispatch(object)
def to_html(o):
    return repr(o)


@dispatch(_strtypes)
def to_html(o):
    return o.replace('\n', '<br>')


@dispatch((object, type, str, unicode), Expr)
def into(a, b, **kwargs):
    result = compute(b, return_type='native', **kwargs)
    kwargs['dshape'] = b.dshape
    return into(a, result, **kwargs)


@dispatch((object, type, str, unicode), _Data)
def into(a, b, **kwargs):
    return into(a, b.data, **kwargs)


@dispatch(_Data, object)
def append(a, b, **kwargs):
    return append(a.data, b, **kwargs)


@dispatch(_Data)
def drop(d):
    return drop(d.data)


def table_length(expr):
    try:
        return expr._len()
    except ValueError:
        return int(expr.count())

_warning_msg = """
In version 0.11, Blaze's expresssion repr will return a standard
representation and will no longer implicitly compute.  Use the `peek()`
method to see a preview of the expression's results.\
"""

use_new_repr = False


def _choose_repr(self):
    if use_new_repr:
        return new_repr(self)
    else:
        warnings.warn(_warning_msg, DeprecationWarning, stacklevel=2)
        return expr_repr(self)


def _warning_repr_html(self):
    if use_new_repr:
        return new_repr(self)
    else:
        warnings.warn(_warning_msg, DeprecationWarning, stacklevel=2)
        return to_html(self)


def new_repr(self):
    fmt = "<`{}` expression; dshape='{}'>"
    return fmt.format(type(self).__name__,
                      sanitized_dshape(self.dshape))


Expr.__repr__ = _choose_repr
Expr.peek = _peek
Expr._repr_html_ = _warning_repr_html
Expr.__len__ = table_length


def intonumpy(data, dtype=None, **kwargs):
    # TODO: Don't ignore other kwargs like copy
    result = odo(data, np.ndarray)
    if dtype and result.dtype != dtype:
        result = result.astype(dtype)
    return result


def convert_base(typ, x):
    x = compute(x)
    try:
        return typ(x)
    except:
        return typ(odo(x, typ))


CORE_SCALAR_TYPES = (float, decimal.Decimal, int, bool, str, Timestamp,
                     datetime.date, datetime.timedelta)
CORE_SEQUENCE_TYPES = (list, dict, tuple, set, Series, DataFrame, np.ndarray)
CORE_TYPES = CORE_SCALAR_TYPES + CORE_SEQUENCE_TYPES


def iscorescalar(x):
    return isinstance(x, CORE_SCALAR_TYPES)


def iscoresequence(x):
    return isinstance(x, CORE_SEQUENCE_TYPES)


def iscoretype(x):
    return isinstance(x, CORE_TYPES)


Expr.__array__ = intonumpy
Expr.__int__ = lambda x: convert_base(int, x)
Expr.__float__ = lambda x: convert_base(float, x)
Expr.__complex__ = lambda x: convert_base(complex, x)
Expr.__bool__ = lambda x: convert_base(bool, x)
Expr.__nonzero__ = lambda x: convert_base(bool, x)
Expr.__iter__ = into(Iterator)



""" MongoDB Backend - Uses aggregation pipeline

If you don't have a mongo server running

    $ conda install mongodb -y
    $ mongod &

>> from blaze import *

>> data = [(1, 'Alice', 100),
...        (2, 'Bob', -200),
...        (3, 'Charlie', 300),
...        (4, 'Denis', 400),
...        (5, 'Edith', -500)]

Migrate data into MongoDB

>> import pymongo
>> db = pymongo.MongoClient().db
>> db.mydata.drop()  # clear out old results
>> _ = into(db.mydata, data, columns=['id', 'name', 'amount'])

Objective: find the name of accounts with negative amount

Using MongoDB query language

>> db.mydata.aggregate([{'$match': {'amount': {'$lt': 0}}}, # doctest: +SKIP
..                      {'$project': {'name': 1, '_id': 0}}])['result']
[{'name': 'Bob'}, {'name': 'Edith'}]

Using Blaze

>> t = data(db.mydata)
>> t[t.amount < 0].name
    name
0    Bob
1  Edith

>> db.mydata.drop()

Uses the aggregation pipeline
http://docs.mongodb.org/manual/core/aggregation-pipeline/
"""

from __future__ import absolute_import, division, print_function

import numbers
import fnmatch

from operator import attrgetter

from pymongo.collection import Collection
from pymongo.database import Database

from datashape.predicates import isscalar
from toolz import pluck, first, get, compose
import toolz
import datetime

from ..expr import (
    And,
    Arithmetic,
    Broadcast,
    By,
    Distinct,
    Eq,
    Expr,
    Field,
    FloorDiv,
    Ge,
    Gt,
    Head,
    Le,
    Lt,
    Ne,
    Or,
    Projection,
    Reduction,
    SimpleSelection,
    Sort,
    Summary,
    Like,
    Symbol,
    ceil,
    count,
    floor,
    math,
    max,
    mean,
    min,
    nelements,
    nunique,
    sum,
    symbol,
)
from ..expr.broadcast import broadcast_collect, Broadcastable
from ..expr.datetime import Day, Month, Year, Minute, Second, UTCFromTimestamp
from ..expr.optimize import simple_selections
from ..compatibility import _strtypes
from ..dispatch import dispatch


__all__ = ['MongoQuery']


@dispatch(Expr, Collection)
def pre_compute(expr, data, scope=None, **kwargs):
    return MongoQuery(data, [])


@dispatch(Expr, Database)
def pre_compute(expr, data, **kwargs):
    return data


class MongoQuery(object):
    """
    A Pair of a pymongo collection and a aggregate query

    We need to carry around both a pymongo collection and a list of
    dictionaries to feed into the aggregation pipeline.  This class
    carries around such a pair.

    Parameters
    ----------
    coll: pymongo.collection.Collection
        A single pymongo collection
    query: list of dicts
        A query to send to coll.aggregate

    >>> q = MongoQuery(db.my_collection, # doctest: +SKIP
    ...     [{'$match': {'name': 'Alice'}},
    ...      {'$project': {'name': 1, 'amount': 1, '_id': 0}}])
    """
    def __init__(self, coll, query):
        self.coll = coll
        self.query = tuple(query)

    def append(self, clause):
        return MongoQuery(self.coll, self.query + (clause,))

    def info(self):
        return self.coll, self.query

    def __eq__(self, other):
        return type(self) == type(other) and self.info() == other.info()

    def __hash__(self):
        return hash((type(self), self.info()))


@dispatch(Expr, (MongoQuery, Collection))
def optimize(expr, seq):
    return broadcast_collect(simple_selections(expr), no_recurse=SimpleSelection)


@dispatch(Head, MongoQuery)
def compute_up(t, q, **kwargs):
    return q.append({'$limit': t.n})


@dispatch(Broadcast, MongoQuery)
def compute_up(t, q, **kwargs):
    s = t._scalars[0]
    d = dict((s[c], symbol(c, s[c].dshape.measure)) for c in s.fields)
    expr = t._scalar_expr._subs(d)
    name = expr._name or 'expr_%d' % abs(hash(expr))
    return q.append({'$project': {name: compute_sub(expr)}})


binops = {'+': 'add',
          '*': 'multiply',
          '/': 'divide',
          '-': 'subtract',
          '%': 'mod'}


def compute_sub(t):
    """
    Build an expression tree in a MongoDB compatible way.

    Parameters
    ----------
    t : Arithmetic
        Scalar arithmetic expression

    Returns
    -------
    sub : dict
        An expression tree

    Examples
    --------
    >>> from blaze import Symbol
    >>> s = symbol('s', 'float64')
    >>> expr = s * 2 + s - 1
    >>> expr
    ((s * 2) + s) - 1
    >>> compute_sub(expr)
    {'$subtract': [{'$add': [{'$multiply': ['$s', 2]}, '$s']}, 1]}

    >>> when = symbol('when', 'datetime')
    >>> compute_sub(s + when.day)
    {'$add': ['$s', {'$dayOfMonth': '$when'}]}
    """
    if isinstance(t, _strtypes + (Symbol,)):
        return '$%s' % t
    elif isinstance(t, numbers.Number):
        return t
    elif isinstance(t, FloorDiv):
        return compute_sub(floor(t.lhs / t.rhs))
    elif isinstance(t, Arithmetic) and hasattr(t, 'symbol') and t.symbol in binops:
        op = binops[t.symbol]
        return {'$%s' % op: [compute_sub(t.lhs), compute_sub(t.rhs)]}
    elif isinstance(t, floor):
        x = compute_sub(t._child)
        return {'$subtract': [x, {'$mod': [x, 1]}]}
    elif isinstance(t, math.abs):
        x = compute_sub(t._child)
        return {'$cond': [{'$lt': [x, 0]},
                          {'$subtract': [0, x]},
                          x]}
    elif isinstance(t, ceil):
        x = compute_sub(t._child)
        return {'$add': [x, {'$subtract': [1, {'$mod': [x, 1]}]}]}
    elif isinstance(t, tuple(datetime_terms)):
        op = datetime_terms[type(t)]
        return {'$%s' % op: compute_sub(t._child)}
    elif isinstance(t, UTCFromTimestamp):
        return {'$add': [datetime.datetime.utcfromtimestamp(0),
                         {'$multiply': [1000, compute_sub(t._child)]}]}
    raise NotImplementedError('Operation %s not supported' % type(t).__name__)


@dispatch((Projection, Field), MongoQuery)
def compute_up(t, q, **kwargs):
    return q.append({'$project': dict((col, 1) for col in t.fields)})


@dispatch(SimpleSelection, MongoQuery)
def compute_up(expr, data, **kwargs):
    predicate = optimize(expr.predicate, data)

    if not isinstance(predicate, Broadcast):
        raise TypeError("Selection predicate must be a broadcastable "
                        "operation.\nReceived an expression of type %r.\n"
                        "Available broadcastable operations are (%s)" %
                        (type(predicate).__name__,
                         ', '.join(map(compose(repr, attrgetter('__name__')),
                                   Broadcastable))))

    s = predicate._scalars[0]
    d = dict((s[c], symbol(c, s[c].dshape.measure)) for c in s.fields)
    expr = predicate._scalar_expr._subs(d)
    return data.append({'$match': match(expr)})


@dispatch(By, MongoQuery)
def compute_up(t, q, **kwargs):
    if not isinstance(t.grouper, (Field, Projection, Symbol)):
        raise ValueError("Complex By operations not supported on MongoDB.\n"
                "The grouping element must be a simple Field or Projection\n"
                "Got %s" % t.grouper)
    apply = optimize(t.apply, q)
    names = apply.fields
    return MongoQuery(q.coll, q.query +
    ({
        '$group': toolz.merge(
                    {'_id': dict((col, '$'+col) for col in t.grouper.fields)},
                    group_apply(apply)
                    )
     },
     {
         '$project': toolz.merge(dict((col, '$_id.'+col) for col in t.grouper.fields),
                                 dict((name, '$' + name) for name in names))
     }))


@dispatch(nunique, MongoQuery)
def compute_up(t, q, **kwargs):
    return MongoQuery(q.coll, q.query +
    ({'$group': {'_id': dict((col, '$'+col) for col in t.fields),
                 t._name: {'$sum': 1}}},
     {'$project': {'_id': 0, t._name: 1}}))


@dispatch(Distinct, MongoQuery)
def compute_up(t, q, **kwargs):
    return MongoQuery(q.coll, q.query +
    ({'$group': {'_id': dict((col, '$'+col) for col in t.fields)}},
     {'$project': toolz.merge(dict((col, '$_id.'+col) for col in t.fields),
                              {'_id': 0})}))


@dispatch(Reduction)
def group_apply(expr):
    """
    Dictionary corresponding to apply part of split-apply-combine operation

    >>> accounts = symbol('accounts', 'var * {name: string, amount: int}')
    >>> group_apply(accounts.amount.sum())
    {'amount_sum': {'$sum': '$amount'}}
    """
    key = expr._name
    col = '$' + expr._child._name
    if isinstance(expr, count):
        return {key: {'$sum': 1}}
    if isinstance(expr, sum):
        return {key: {'$sum': col}}
    if isinstance(expr, max):
        return {key: {'$max': col}}
    if isinstance(expr, min):
        return {key: {'$min': col}}
    if isinstance(expr, mean):
        return {key: {'$avg': col}}
    raise NotImplementedError("Reduction %s not yet supported in MongoDB"
                              % type(expr).__name__)


reductions = {mean: '$avg', count: '$sum', max: '$max', min: '$min', sum: '$sum'}


def scalar_expr(expr):
    if isinstance(expr, Broadcast):
        s = expr._scalars[0]
        d = dict((s[c], symbol(c, s[c].dshape.measure)) for c in s.fields)
        return expr._scalar_expr._subs(d)
    elif isinstance(expr, Field):
        return symbol(expr._name, expr.dshape.measure)
    else:
        # TODO: This is a hack
        # broadcast_collect should reach into summary, By, selection
        # And perform these kinds of optimizations itself
        expr2 = broadcast_collect(expr)
        if not expr2.isidentical(expr):
            return scalar_expr(expr2)


@dispatch(Summary)
def group_apply(expr):
    # TODO: implement columns variable more generally when Broadcast works
    reducs = expr.values
    names = expr.fields
    values = [(name, c, getattr(c._child, 'column', None) or name)
               for name, c in zip(names, reducs)]

    query = dict((k, {reductions[type(v)]: 1 if isinstance(v, count)
                                        else compute_sub(scalar_expr(v._child))})
                 for k, v, z in values)
    return query


@dispatch((count, nelements), MongoQuery)
def compute_up(t, q, **kwargs):
    name = t._name
    return q.append({'$group': {'_id': {}, name: {'$sum': 1}}})


@dispatch((sum, min, max, mean), MongoQuery)
def compute_up(t, q, **kwargs):
    name = t._name
    reduction = {sum: '$sum', min: '$min', max: '$max', mean: '$avg'}[type(t)]
    column = '$' + t._child._name
    arg = {'$group': {'_id': {}, name: {reduction: column}}}
    return q.append(arg)


@dispatch(Sort, MongoQuery)
def compute_up(t, q, **kwargs):
    return q.append({'$sort': {t.key: 1 if t.ascending else -1}})


datetime_terms = {Day: 'dayOfMonth',
                  Month: 'month',
                  Year: 'year',
                  Minute: 'minute',
                  Second: 'second'}


@dispatch(Field, Database)
def compute_up(expr, data, **kwargs):
    return getattr(data, expr._name)


@dispatch(Expr, Collection)
def post_compute(e, c, scope=None):
    """
    Calling compute on a raw collection?  Compute on an empty MongoQuery.
    """
    return post_compute(e, MongoQuery(c, ()), scope=scope)


def get_result(result):
    try:
        return result['result']
    except TypeError:
        return list(result)


@dispatch(Expr, MongoQuery)
def post_compute(e, q, scope=None):
    """
    Execute a query using MongoDB's aggregation pipeline

    The compute_up functions operate on Mongo Collection / list-of-dict
    queries.  Once they're done we need to actually execute the query on
    MongoDB.  We do this using the aggregation pipeline framework.

    http://docs.mongodb.org/manual/core/aggregation-pipeline/
    """
    scope = {'$project': toolz.merge({'_id': 0},  # remove mongo identifier
                                 dict((col, 1) for col in e.fields))}
    q = q.append(scope)

    if not e.dshape.shape:  # not a collection
        result = get_result(q.coll.aggregate(list(q.query)))[0]
        if isscalar(e.dshape.measure):
            return result[e._name]
        else:
            return get(e.fields, result)

    dicts = get_result(q.coll.aggregate(list(q.query)))

    if isscalar(e.dshape.measure):
        return list(pluck(e.fields[0], dicts, default=None))  # dicts -> values
    else:
        return list(pluck(e.fields, dicts, default=None))  # dicts -> tuples


@dispatch(Broadcast, MongoQuery)
def post_compute(e, q, scope=None):
    """Compute the result of a Broadcast expression.
    """
    columns = dict((col, 1) for qry in q.query
                   for col in qry.get('$project', []))
    scope = {'$project': toolz.merge({'_id': 0},  # remove mongo identifier
                                 dict((col, 1) for col in columns))}
    q = q.append(scope)
    dicts = get_result(q.coll.aggregate(list(q.query)))

    assert len(columns) == 1
    return list(pluck(first(columns.keys()), dicts))


def name(e):
    """

    >>> name(Symbol('x', 'int32'))
    'x'
    >>> name(1)
    1
    """
    if isinstance(e, Symbol):
        return e._name
    elif isinstance(e, Expr):
        raise NotImplementedError("Complex queries not yet supported")
    else:
        return e


opnames = {
    Lt: '$lt',
    Le: '$lte',
    Gt: '$gt',
    Ge: '$gte',
    Ne: '$ne',
    And: '$and',
    Or: '$or',
    Like: '$regex',
}


@dispatch((Lt, Le, Gt, Ge, Ne))
def match(expr):
    return {name(expr.lhs): {opnames[type(expr)]: expr.rhs}}


@dispatch((And, Or))
def match(expr):
    return {opnames[type(expr)]: [match(expr.lhs), match(expr.rhs)]}


@dispatch(Eq)
def match(expr):
    return {name(expr.lhs): name(expr.rhs)}


@dispatch(Like)
def match(expr):
    return {
        name(expr._child): {
            opnames[type(expr)]: fnmatch.translate(expr.pattern)
        }
    }


@dispatch(object)
def match(expr):
    """ Match query for MongoDB

    Examples
    --------
    >>> x = symbol('x', 'int32')
    >>> name = symbol('name', 'string')
    >>> match(name == 'Alice')
    {'name': 'Alice'}
    >>> match(x > 10)
    {'x': {'$gt': 10}}
    >>> match(10 > x)
    {'x': {'$lt': 10}}
    >>> match((x > 10) & (name == 'Alice'))  # doctest: +SKIP
    {'x': {'$gt': 10}, 'name': 'Alice'}
    >>> match((x > 10) | (name == 'Alice'))
    {'$or': [{'x': {'$gt': 10}}, {'name': 'Alice'}]}
    """
    raise NotImplementedError(
        'Matching not supported on expressions of type %r' % type(expr).__name__
    )



from __future__ import absolute_import, division, print_function

import os
from blaze.utils import tmpfile


def test_tmpfile():
    with tmpfile() as f:
        with open(f, 'w') as a:
            a.write('')
        with tmpfile() as g:
            assert f != g

    assert not os.path.exists(f)



from blaze.expr import symbol
import blaze.expr.datetime as bdt
from blaze.compatibility import builtins
from datashape import dshape

import pytest


def test_datetime_dshape():
    t = symbol('t', '5 * {name: string, when: datetime}')
    assert t.when.day.dshape == dshape('5 * int64')
    assert t.when.date.dshape == dshape('5 * date')


def test_date_attribute():
    t = symbol('t', '5 * {name: string, when: datetime}')
    expr = t.when.day
    assert eval(str(expr)).isidentical(expr)


def test_invalid_date_attribute():
    t = symbol('t', '5 * {name: string, when: datetime}')
    with pytest.raises(AttributeError):
        t.name.day


def test_date_attribute_completion():
    t = symbol('t', '5 * {name: string, when: datetime}')
    assert 'day' in dir(t.when)
    assert 'day' not in dir(t.name)
    assert not builtins.all([x.startswith('__') and x.endswith('__')
                            for x in dir(t.name)])


def test_datetime_attribute_name():
    t = symbol('t', '5 * {name: string, when: datetime}')
    assert 'when' in t.when.day._name


def test_utcfromtimestamp():
    t = symbol('t', '10 * {name: string, ts: int32}')
    assert t.ts.utcfromtimestamp.dshape == dshape('10 * datetime')


def test_isdatelike():
    assert not bdt.isdatelike('int32')
    assert bdt.isdatelike('?date')
    assert not bdt.isdatelike('{is_outdated: bool}')


def test_truncate_names():
    t = symbol('t', '5 * {name: string, when: datetime}')
    assert t.when.truncate(days=2)._name == 'when'


def test_truncate_repr():
    t = symbol('t', '5 * {name: string, when: datetime}')
    assert str(t.when.truncate(days=2)) == 't.when.truncate(days=2)'
    assert str(t.when.date.truncate(month=3)) == 't.when.date.truncate(months=3)'
    assert str(t.when.date.truncate(ns=4)) == 't.when.date.truncate(nanoseconds=4)'
    assert str(t.when.date.truncate(days=4.2)) == 't.when.date.truncate(days=4.2)'
    assert str(t.when.date.truncate(days=4.2876)) == 't.when.date.truncate(days=4.2876)'
    assert str(t.when.date.truncate(days=4.287654)) == 't.when.date.truncate(days=4.28765)'


def test_truncate_raises_with_no_arguments():
    t = symbol('t', '5 * {name: string, when: datetime}')
    with pytest.raises(TypeError):
        t.when.truncate()


@pytest.mark.parametrize('attr',
                         ['date', 'year', 'month', 'day', 'time', 'hour',
                          'second', 'millisecond', 'microsecond'])
def test_attributes(attr):
    t = symbol('t', 'var * datetime')
    assert getattr(t, attr).dshape is not None
    assert getattr(t, attr)._child is t


def test_dt_namespace():
    t = symbol('t', 'var * {when: datetime}')
    assert bdt.Ceil(t.when, 's').isidentical(t.when.dt.ceil('s'))
    assert bdt.date(t.when).isidentical(t.when.dt.date())
    assert bdt.day(t.when).isidentical(t.when.dt.day())
    assert bdt.dayofweek(t.when).isidentical(t.when.dt.dayofweek())
    assert bdt.dayofyear(t.when).isidentical(t.when.dt.dayofyear())
    assert bdt.days_in_month(t.when).isidentical(t.when.dt.days_in_month())
    assert bdt.Floor(t.when, 's').isidentical(t.when.dt.floor('s'))
    assert bdt.hour(t.when).isidentical(t.when.dt.hour())
    assert bdt.is_month_end(t.when).isidentical(t.when.dt.is_month_end())
    assert bdt.is_month_start(t.when).isidentical(t.when.dt.is_month_start())
    assert bdt.is_quarter_end(t.when).isidentical(t.when.dt.is_quarter_end())
    assert bdt.is_quarter_start(t.when).isidentical(t.when.dt.is_quarter_start())
    assert bdt.is_year_end(t.when).isidentical(t.when.dt.is_year_end())
    assert bdt.is_year_start(t.when).isidentical(t.when.dt.is_year_start())
    assert bdt.microsecond(t.when).isidentical(t.when.dt.microsecond())
    assert bdt.millisecond(t.when).isidentical(t.when.dt.millisecond())
    assert bdt.minute(t.when).isidentical(t.when.dt.minute())
    assert bdt.month(t.when).isidentical(t.when.dt.month())
    assert bdt.nanosecond(t.when).isidentical(t.when.dt.nanosecond())
    assert bdt.Round(t.when, 's').isidentical(t.when.dt.round('s'))
    assert bdt.quarter(t.when).isidentical(t.when.dt.quarter())
    assert bdt.second(t.when).isidentical(t.when.dt.second())
    assert bdt.strftime(t.when, 'format').isidentical(t.when.dt.strftime('format'))
    assert bdt.time(t.when).isidentical(t.when.dt.time())
    assert bdt.truncate(t.when, days=2).isidentical(t.when.dt.truncate(days=2))
    assert bdt.week(t.when).isidentical(t.when.dt.week())
    assert bdt.weekday(t.when).isidentical(t.when.dt.weekday())
    assert bdt.weekofyear(t.when).isidentical(t.when.dt.weekofyear())
    assert bdt.year(t.when).isidentical(t.when.dt.year())

def test_td_namespace():
    t = symbol('t', 'var * {span: timedelta}')
    assert bdt.nanoseconds(t.span).isidentical(t.span.dt.nanoseconds())
    assert bdt.seconds(t.span).isidentical(t.span.dt.seconds())
    assert bdt.total_seconds(t.span).isidentical(t.span.dt.total_seconds())




from __future__ import print_function, division, absolute_import

import numpy as np
import pandas as pd

from datashape import dshape, discover, integral, floating, boolean, complexes
from datashape.predicates import isscalar, isrecord, iscollection

from .dispatch import dispatch
from .expr import Expr, Field, ndim
from .compute import compute
from .compatibility import unicode

from odo import odo


class CachedDataset(object):
    __slots__ = 'data', 'cache'

    def __init__(self, data, cache=None):
        self.data = data
        self.cache = cache if cache is not None else {}


@dispatch(CachedDataset)
def discover(d, **kwargs):
    return discover(d.data, **kwargs)


@dispatch(Field, CachedDataset)
def compute_up(expr, data, **kwargs):
    return data.data[expr._name]


@dispatch(Expr, CachedDataset)
def compute_down(expr, data, **kwargs):
    try:
        return data.cache[expr]
    except KeyError:
        pass

    leaf, = expr._leaves()

    # Do work
    result = compute(expr, {leaf: data.data}, **kwargs)

    ds = expr.dshape
    new_type = concrete_type(ds)

    # If the result is not the concrete data type for the datashape then make
    # it concrete
    if not isinstance(result, new_type):
        result = odo(result, new_type, dshape=ds)

    # Cache result
    data.cache[expr] = result

    return result


def concrete_type(ds):
    """A type into which we can safely deposit streaming data.

    Parameters
    ----------
    ds : DataShape

    Returns
    -------
    type : type
        The concrete type corresponding to the DataShape `ds`

    Notes
    -----
    * This will return a Python type if possible
    * Option types are not handled specially. The base type of the option type
      is returned.

    Examples
    --------
    >>> concrete_type('5 * int')
    <class 'pandas.core.series.Series'>
    >>> concrete_type('var * {name: string, amount: int}')
    <class 'pandas.core.frame.DataFrame'>
    >>> concrete_type('float64')
    <... 'float'>
    >>> concrete_type('float32')
    <... 'float'>
    >>> concrete_type('int64')
    <... 'int'>
    >>> concrete_type('int32')
    <... 'int'>
    >>> concrete_type('uint8')
    <... 'int'>
    >>> concrete_type('bool')
    <... 'bool'>
    >>> concrete_type('complex[float64]')
    <... 'complex'>
    >>> concrete_type('complex[float32]')
    <... 'complex'>
    >>> concrete_type('?int64')
    <... 'int'>
    """
    if isinstance(ds, (str, unicode)):
        ds = dshape(ds)
    if not iscollection(ds) and isscalar(ds.measure):
        measure = getattr(ds.measure, 'ty', ds.measure)
        if measure in integral.types:
            return int
        elif measure in floating.types:
            return float
        elif measure in boolean.types:
            return bool
        elif measure in complexes.types:
            return complex
        else:
            return ds.measure.to_numpy_dtype().type
    if not iscollection(ds):
        return type(ds)
    if ndim(ds) == 1:
        return pd.DataFrame if isrecord(ds.measure) else pd.Series
    if ndim(ds) > 1:
        return np.ndarray
    return list



from __future__ import absolute_import, division, print_function

from collections import Iterable

from datashape import DataShape
from odo.utils import copydoc

from .expressions import Expr, ndim, symbol

__all__ = 'Transpose', 'TensorDot', 'dot', 'transpose', 'tensordot'


class Transpose(Expr):
    """ Transpose dimensions in an N-Dimensional array

    Examples
    --------

    >>> x = symbol('x', '10 * 20 * int32')
    >>> x.T
    transpose(x)
    >>> x.T.shape
    (20, 10)

    Specify axis ordering with axes keyword argument

    >>> x = symbol('x', '10 * 20 * 30 * int32')
    >>> x.transpose([2, 0, 1])
    transpose(x, axes=[2, 0, 1])
    >>> x.transpose([2, 0, 1]).shape
    (30, 10, 20)
    """
    _arguments = '_child', 'axes'

    def _dshape(self):
        s = self._child.shape
        return DataShape(*(tuple([s[i] for i in self.axes]) +
                           (self._child.dshape.measure,)))

    def __str__(self):
        if self.axes == tuple(range(ndim(self)))[::-1]:
            return 'transpose(%s)' % self._child
        else:
            return 'transpose(%s, axes=%s)' % (self._child, list(self.axes))


@copydoc(Transpose)
def transpose(expr, axes=None):
    if axes is None:
        axes = tuple(range(ndim(expr)))[::-1]
    if isinstance(axes, list):
        axes = tuple(axes)
    return Transpose(expr, axes)


@copydoc(Transpose)
def T(expr):
    return transpose(expr)


class TensorDot(Expr):
    """ Dot Product: Contract and sum dimensions of two arrays

    >>> x = symbol('x', '20 * 20 * int32')
    >>> y = symbol('y', '20 * 30 * int32')

    >>> x.dot(y)
    tensordot(x, y)

    >>> tensordot(x, y, axes=[0, 0])
    tensordot(x, y, axes=[0, 0])
    """
    _arguments = 'lhs', 'rhs', '_left_axes', '_right_axes'
    __inputs__ = 'lhs', 'rhs'

    def _dshape(self):
        # Compute shape
        shape = tuple([d for i, d in enumerate(self.lhs.shape)
                         if i not in self._left_axes] +
                      [d for i, d in enumerate(self.rhs.shape)
                         if i not in self._right_axes])

        # Compute measure by mimicking a mul and add
        l = symbol('l', self.lhs.dshape.measure)
        r = symbol('r', self.rhs.dshape.measure)
        measure = ((l * r) + (l * r)).dshape.measure

        return DataShape(*(shape + (measure,)))

    def __str__(self):
        if self.isidentical(tensordot(self.lhs, self.rhs)):
            return 'tensordot(%s, %s)' % (self.lhs, self.rhs)
        else:
            la = self._left_axes
            if len(la) == 1:
                la = la[0]
            ra = self._right_axes
            if len(ra) == 1:
                ra = ra[0]
            return 'tensordot(%s, %s, axes=[%s, %s])' % (
                self.lhs, self.rhs, str(la), str(ra)
            )


@copydoc(TensorDot)
def tensordot(lhs, rhs, axes=None):
    if axes is None:
        left = ndim(lhs) - 1
        right = 0
    elif isinstance(axes, Iterable):
        left, right = axes
    else:
        left, right = axes, axes

    if isinstance(left, int):
        left = (left,)
    if isinstance(right, int):
        right = (right,)
    if isinstance(left, list):
        left = tuple(left)
    if isinstance(right, list):
        right = tuple(right)

    return TensorDot(lhs, rhs, left, right)


@copydoc(TensorDot)
def dot(lhs, rhs):
    return tensordot(lhs, rhs)


from datashape.predicates import isnumeric, isboolean
from .expressions import dshape_method_list, method_properties

dshape_method_list.extend([
    (lambda ds: ndim(ds) > 1, set([transpose])),
    (lambda ds: ndim(ds) == 2, set([T])),
    (lambda ds: ndim(ds) >= 1 and (isnumeric(ds) or isboolean(ds)), set([dot]))
    ])

method_properties.add(T)



from __future__ import absolute_import, division, print_function

from weakref import WeakKeyDictionary

from toolz import curry, concat, first, memoize
from multipledispatch import MDNotImplementedError

from ..expr import (
    Distinct,
    ElemWise,
    Expr,
    Field,
    Head,
    Projection,
    Slice,
    Symbol,
    path,
    symbol,
)
from ..expr.optimize import lean_projection, simple_selections
from ..expr.split import split
from ..partition import partitions
from .core import compute
from .pmap import get_default_pmap

from collections import Iterator, Iterable
import datashape
import bcolz
import numpy as np
import pandas as pd


from ..dispatch import dispatch
from odo import into

__all__ = ['bcolz']

COMFORTABLE_MEMORY_SIZE = 1e9


@memoize(cache=WeakKeyDictionary())
def box(type_):
    """Create a non-iterable box type for an object.

    Parameters
    ----------
    type_ : type
        The type to create a box for.

    Returns
    -------
    box : type
        A type to box values of type ``type_``.
    """
    class c(object):
        __slots__ = 'value',

        def __init__(self, value):
            if not isinstance(value, type_):
                raise TypeError(
                    "values must be of type '%s' (received '%s')" % (
                        type_.__name__, type(value).__name__,
                    ),
                )
            self.value = value

    c.__name__ = 'box(%s)' % type_.__name__
    return c


@dispatch(Expr, (box(bcolz.ctable), box(bcolz.carray)))
def optimize(expr, _):
    return simple_selections(lean_projection(expr))


@dispatch(Expr, (bcolz.ctable, bcolz.carray))
def pre_compute(expr, data, scope=None, **kwargs):
    # box the data so that we don't need to deal with ambiguity of ctable
    # and carray being instances of the Iterator ABC.
    return box(type(data))(data)


@dispatch(Expr, (box(bcolz.ctable), box(bcolz.carray)))
def post_compute(expr, data, **kwargs):
    # Unbox the bcolz objects.
    return data.value


@dispatch((box(bcolz.carray), box(bcolz.ctable)))
def discover(data):
    val = data.value
    return datashape.from_numpy(val.shape, val.dtype)

Cheap = (Head, ElemWise, Distinct, Symbol)


@dispatch(Head, (box(bcolz.ctable), box(bcolz.carray)))
def compute_down(expr, data, **kwargs):
    """ Cheap and simple computation in simple case

    If we're given a head and the entire expression is cheap to do (e.g.
    elemwises, selections, ...) then compute on data directly, without
    parallelism"""
    leaf = expr._leaves()[0]
    if all(isinstance(e, Cheap) for e in path(expr, leaf)):
        val = data.value
        return compute(
            expr,
            {leaf: into(Iterator, val)},
            return_type='native',
            **kwargs
        )
    else:
        raise MDNotImplementedError()


@dispatch(Field, box(bcolz.ctable))
def compute_up(expr, data, **kwargs):
    return data.value[str(expr._name)]


@dispatch(Projection, box(bcolz.ctable))
def compute_up(expr, data, **kwargs):
    return data.value[list(map(str, expr.fields))]


@dispatch(Slice, (box(bcolz.carray), box(bcolz.ctable)))
def compute_up(expr, data, **kwargs):
    return data.value[expr.index]


def compute_chunk(source, chunk, chunk_expr, data_index):
    part = source[data_index]
    return compute(chunk_expr, {chunk: part}, return_type='native')


def get_chunksize(data):
    if isinstance(data, bcolz.carray):
        return data.chunklen
    elif isinstance(data, bcolz.ctable):
        return min(data[c].chunklen for c in data.names)
    else:
        raise TypeError("Don't know how to compute chunksize for type %r" %
                        type(data).__name__)


@dispatch(Expr, (box(bcolz.carray), box(bcolz.ctable)))
def compute_down(expr, data, chunksize=None, map=None, **kwargs):
    data = data.value
    if map is None:
        map = get_default_pmap()

    leaf = expr._leaves()[0]

    if chunksize is None:
        chunksize = max(2**16, get_chunksize(data))

    # If the bottom expression is a projection or field then want to do
    # compute_up first
    children = {
        e for e in expr._traverse()
        if isinstance(e, Expr)
        and any(i is expr._leaves()[0] for i in e._inputs)
    }
    if len(children) == 1 and isinstance(first(children), (Field, Projection)):
        raise MDNotImplementedError()

    chunk = symbol('chunk', chunksize * leaf.schema)
    (chunk, chunk_expr), (agg, agg_expr) = split(leaf, expr, chunk=chunk)

    data_parts = partitions(data, chunksize=(chunksize,))

    parts = list(map(curry(compute_chunk, data, chunk, chunk_expr),
                     data_parts))

    if isinstance(parts[0], np.ndarray):
        intermediate = np.concatenate(parts)
    elif isinstance(parts[0], pd.DataFrame):
        intermediate = pd.concat(parts)
    elif isinstance(parts[0], Iterable):
        intermediate = list(concat(parts))
    else:
        raise TypeError("Don't know how to concatenate objects of type %r" %
                        type(parts[0]).__name__)

    return compute(agg_expr, {agg: intermediate}, return_type='native')


def _asarray(a):
    if isinstance(a, (bcolz.carray, bcolz.ctable)):
        return a[:]
    return np.array(list(a))


@compute_down.register(Expr, (box(bcolz.carray), box(bcolz.ctable)), Iterable)
@compute_down.register(Expr, Iterable, (box(bcolz.carray), box(bcolz.ctable)))
def bcolz_mixed(expr, a, b, **kwargs):
    return compute(
        expr,
        dict(zip(expr._leaves(), map(_asarray, (a.value, b.value)))),
        return_type='native',
    )



from __future__ import absolute_import, division, print_function

from .serialization import *
from .json_dumps import json_dumps
from .json_dumps_trusted import json_dumps_trusted
from .object_hook import object_hook
from .object_hook_trusted import object_hook_trusted



from blaze import into, drop, create_index
from blaze.compute.core import compute_down, compute_up
from multipledispatch.conflict import ambiguities
from blaze.expr.arithmetic import scalar_coerce


def test_no_dispatch_ambiguities():
    for func in [into, compute_up, compute_down, drop, create_index,
                 scalar_coerce]:
        assert not ambiguities(func.funcs)



from __future__ import absolute_import, division, print_function

from functools import partial
import numpy as np
import tables as tb
from datashape import Record, from_numpy, datetime_, date_

from blaze.expr import (Selection, Head, Field, Broadcast, Projection, Symbol,
                        Sort, Reduction, count, Slice, Expr, nelements,
                        UnaryOp, BinOp)
from blaze.compatibility import basestring, map
from blaze.expr.optimize import simple_selections
from ..dispatch import dispatch
from .numexpr import broadcast_numexpr_collect, print_numexpr


__all__ = ['drop', 'create_index']


@dispatch(tb.Table)
def discover(t):
    return t.shape[0] * Record([[col, discover(getattr(t.cols, col))]
                                for col in t.colnames])


@dispatch(tb.Column)
def discover(c):
    dshape = from_numpy(c.shape, c.dtype)
    return {'time64': datetime_, 'time32': date_}.get(c.type,
                                                      dshape.subshape[1])


@dispatch(tb.Table)
def drop(t):
    t.remove()


@dispatch(tb.Table, basestring)
def create_index(t, column, name=None, **kwargs):
    create_index(getattr(t.cols, column), **kwargs)


@dispatch(tb.Table, (list, tuple))
def create_index(t, columns, name=None, **kwargs):
    if not all(map(partial(hasattr, t.cols), columns)):
        raise ValueError('table %s does not have all passed in columns %s' %
                         (t, columns))
    for column in columns:
        create_index(t, column, **kwargs)


@dispatch(tb.Column)
def create_index(c, optlevel=9, kind='full', name=None, **kwargs):
    c.create_index(optlevel=optlevel, kind=kind, **kwargs)


@dispatch(Selection, tb.Table)
def compute_up(expr, data, **kwargs):
    predicate = optimize(expr.predicate, data)
    assert isinstance(predicate, Broadcast)

    s = predicate._scalars[0]
    cols = [s[field] for field in s.fields]
    expr_str = print_numexpr(cols, predicate._scalar_expr)
    return data.read_where(expr_str)


@dispatch(Symbol, tb.Table)
def compute_up(ts, t, **kwargs):
    return t


@dispatch(Reduction, (tb.Column, tb.Table))
def compute_up(r, c, **kwargs):
    return compute_up(r, c[:])


@dispatch(Projection, tb.Table)
def compute_up(proj, t, **kwargs):
    # only options here are
    # read the whole thing in and then select
    # or
    # create an output array that is at most the size of the on disk table and
    # fill it will the columns iteratively
    # both of these options aren't ideal but pytables has no way to select
    # multiple column subsets so pick the one where we can optimize for the best
    # case rather than prematurely pessimizing
    #
    # TODO: benchmark on big tables because i'm not sure exactly what the
    # implications here are for memory usage
    columns = proj.fields
    dtype = np.dtype([(col, t.dtype[col]) for col in columns])
    out = np.empty(t.shape, dtype=dtype)
    for c in columns:
        out[c] = t.col(c)
    return out


@dispatch(Field, tb.File)
def compute_up(expr, data, **kwargs):
    return data.getNode('/')._v_children[expr._name]


@dispatch(Field, tb.Node)
def compute_up(expr, data, **kwargs):
    return data._v_children[expr._name]


@dispatch(Field, tb.Table)
def compute_up(c, t, **kwargs):
    return getattr(t.cols, c._name)


@dispatch(count, tb.Column)
def compute_up(r, c, **kwargs):
    return len(c)


@dispatch(Head, (tb.Column, tb.Table))
def compute_up(h, t, **kwargs):
    return t[:h.n]


@dispatch(Broadcast, tb.Table)
def compute_up(expr, data, **kwargs):
    if len(expr._children) != 1:
        raise ValueError("Only one child in Broadcast allowed")

    s = expr._scalars[0]
    cols = [s[field] for field in s.fields]
    expr_str = print_numexpr(cols, expr._scalar_expr)
    uservars = dict((c, getattr(data.cols, c)) for c in s.fields)
    e = tb.Expr(expr_str, uservars=uservars, truediv=True)
    return e.eval()


@dispatch(Sort, tb.Table)
def compute_up(s, t, **kwargs):
    if isinstance(s.key, Field) and s.key._child.isidentical(s._child):
        key = s.key._name
    else:
        key = s.key
    assert hasattr(t.cols, key), 'Table has no column(s) %s' % s.key
    result = t.read_sorted(sortby=key, checkCSI=True)
    if s.ascending:
        return result
    return result[::-1]


@dispatch(Slice, (tb.Table, tb.Column))
def compute_up(expr, x, **kwargs):
    return x[expr.index]


Broadcastable = UnaryOp, BinOp, Field
WantToBroadcast = UnaryOp, BinOp


@dispatch(Expr, tb.Table)
def optimize(expr, seq):
    return broadcast_numexpr_collect(
        simple_selections(expr),
        broadcastable=Broadcastable,
        want_to_broadcast=WantToBroadcast,
        no_recurse=Selection,
    )


@dispatch(nelements, tb.Table)
def compute_up(expr, x, **kwargs):
    return compute_up.dispatch(type(expr), np.ndarray)(expr, x, **kwargs)



from blaze import compute, data, symbol, discover
from blaze.utils import example


flag = [False]

def mymap(func, *args):
    flag[0] = True
    return map(func, *args)


def test_map_called_on_data_star():
    r = data(example('accounts_*.csv'))
    s = symbol('s', discover(r))
    flag[0] = False
    a = compute(s.count(), r)
    b = compute(s.count(), r, map=mymap)
    assert a == b
    assert flag[0]



from blaze.expr import *
from blaze.expr.arrays import *
from blaze.utils import raises


w = symbol('w', '5 * 6 * int32')
x = symbol('x', '4 * 5 * int32')
y = symbol('y', '5 * int32')
z = symbol('z', '4 * int32')

def test_shape_of_binop_of_differently_shaped_arrays():
    assert (x + y).shape == x.shape


def test_transpose():
    assert ndim(x.dshape) == 2
    assert x.T.shape == (5, 4)


def test_default_transpose_axes():
    a = symbol('a', '10 * 20 * 30 * 40 * int32')
    assert transpose(a).shape == a.shape[::-1]


def test_dot():
    assert x.dot(y).shape == (4,)
    assert x.dot(w).shape == (4, 6)


def test_tensordot():
    expr = tensordot(x, z, axes=[0, 0])
    assert expr.shape == (5,)

    expr = tensordot(x, z, axes=[[0], [0]])
    assert expr.shape == (5,)


def test_tensordot_schema_matches_mul():
    x = symbol('x', '5 * int64')
    y = symbol('y', '5 * float32')

    assert tensordot(x, y).dshape.measure == (x + y).dshape.measure


def test_shapes_raise_errors():
    assert raises(ValueError, lambda: w + x)



from __future__ import absolute_import, division, print_function


import pytest
h5py = pytest.importorskip('h5py')

import sys
from distutils.version import LooseVersion

import numpy as np

from datashape import discover

from blaze import compute
from blaze.expr import symbol
from blaze.utils import tmpfile
from blaze.compute.h5py import pre_compute, optimize


def eq(a, b):
    return (a == b).all()


x = np.arange(20*24, dtype='f4').reshape((20, 24))


@pytest.yield_fixture
def file():
    with tmpfile('.h5') as filename:
        f = h5py.File(filename)
        d = f.create_dataset('/x', shape=x.shape, dtype=x.dtype,
                             fillvalue=0.0, chunks=(4, 6))
        d[:] = x
        yield f
        f.close()


@pytest.yield_fixture
def data(file):
    yield file['/x']


@pytest.yield_fixture
def data_1d_chunks():
    with tmpfile('.h5') as filename:
        f = h5py.File(filename)
        d = f.create_dataset('/x', shape=x.shape, dtype=x.dtype,
                             fillvalue=0.0, chunks=(1, 24))
        d[:] = x

        yield d

        f.close()


rec = np.empty(shape=(20, 24), dtype=[('x', 'i4'), ('y', 'i4')])
rec['x'] = 1
rec['y'] = 2


@pytest.yield_fixture
def recdata():
    with tmpfile('.h5') as filename:
        f = h5py.File(filename)
        d = f.create_dataset('/x', shape=rec.shape,
                             dtype=rec.dtype,
                             chunks=(4, 6))
        d['x'] = rec['x']
        d['y'] = rec['y']
        yield d
        f.close()

s = symbol('s', discover(x))

two_point_five_and_windows_py3 = \
    pytest.mark.skipif(sys.platform == 'win32' and
                       h5py.__version__ == LooseVersion('2.5.0') and
                       sys.version_info[0] == 3,
                       reason=('h5py 2.5.0 issue with varlen string types: '
                               'https://github.com/h5py/h5py/issues/593'))



def test_slicing(data):
    for idx in [0, 1, (0, 1), slice(1, 3), (0, slice(1, 5, 2))]:
        assert eq(compute(s[idx], data), x[idx])


def test_reductions(data):
    assert eq(compute(s.sum(), data), x.sum())
    assert eq(compute(s.sum(axis=1), data), x.sum(axis=1))
    assert eq(compute(s.sum(axis=0), data), x.sum(axis=0))
    assert eq(compute(s[0], data), x[0])
    assert eq(compute(s[-1], data), x[-1])


@two_point_five_and_windows_py3
def test_mixed(recdata):
    s = symbol('s', discover(recdata))
    expr = (s.x + 1).sum(axis=1)
    assert eq(compute(expr, recdata), compute(expr, rec))


def test_uneven_chunk_size(data):
    assert eq(compute(s.sum(axis=1), data, chunksize=(7, 7)),
              x.sum(axis=1))


@two_point_five_and_windows_py3
def test_nrows_3D_records(recdata):
    s = symbol('s', discover(recdata))
    assert not hasattr(s, 'nrows')


@pytest.mark.xfail(raises=AttributeError,
                   reason="We don't support nrows on arrays")
def test_nrows_array(data):
    assert compute(s.nrows, data) == len(data)


@two_point_five_and_windows_py3
def test_nelements_records(recdata):
    s = symbol('s', discover(recdata))
    assert compute(s.nelements(), recdata) == np.prod(recdata.shape)
    np.testing.assert_array_equal(compute(s.nelements(axis=0), recdata),
                                  np.zeros(recdata.shape[1]) + recdata.shape[0])


def test_nelements_array(data):
    lhs = compute(s.nelements(axis=1), data)
    rhs = data.shape[1]
    np.testing.assert_array_equal(lhs, rhs)

    lhs = compute(s.nelements(axis=0), data)
    rhs = data.shape[0]
    np.testing.assert_array_equal(lhs, rhs)

    lhs = compute(s.nelements(axis=(0, 1)), data)
    rhs = np.prod(data.shape)
    np.testing.assert_array_equal(lhs, rhs)


def test_field_access_on_file(file):
    s = symbol('s', '{x: 20 * 24 * float32}')
    d = compute(s.x, file)
    # assert isinstance(d, h5py.Dataset)
    assert eq(d[:], x)


def test_field_access_on_group(file):
    s = symbol('s', '{x: 20 * 24 * float32}')
    d = compute(s.x, file['/'])
    # assert isinstance(d, h5py.Dataset)
    assert eq(d[:], x)


def test_compute_on_file(file):
    s = symbol('s', discover(file))

    assert eq(compute(s.x.sum(axis=1), file),
              x.sum(axis=1))

    assert eq(compute(s.x.sum(), file, chunksize=(4, 6)),
              x.sum())


def test_compute_on_1d_chunks(data_1d_chunks):
    assert eq(compute(s.sum(), data_1d_chunks),
              x.sum())


def test_arithmetic_on_small_array(data):
    s = symbol('s', discover(data))

    assert eq(compute(s + 1, data),
              compute(s + 1, x))


def test_arithmetic_on_small_array_from_file(file):
    """ Want to make sure that we call pre_compute on Dataset
        Even when it's not the leaf data input. """
    s = symbol('s', discover(file))

    assert eq(compute(s.x + 1, file),
              x + 1)


def test_pre_compute_doesnt_collapse_slices(data):
    s = symbol('s', discover(data))
    assert pre_compute(s[:5], data) is data


def test_optimize_slicing(data):
    a = symbol('a', discover(data))
    b = symbol('b', discover(data))
    assert optimize((a + 1)[:3], data).isidentical(a[:3] + 1)

    assert optimize((a + b)[:3], data).isidentical(a[:3] + b[:3])


def test_optimize_slicing_on_file(file):
    f = symbol('f', discover(file))
    assert optimize((f.x + 1)[:5], file).isidentical(f.x[:5] + 1)


def test_arithmetic_and_then_slicing(data):
    s = symbol('s', discover(data))

    assert eq(compute((2*s + 1)[0], data, pre_compute=False),
              2*x[0] + 1)



from __future__ import absolute_import, division, print_function
from collections import Mapping

import math
import itertools
import operator
import pytest
from datetime import datetime, date
import datashape
from datashape.py2help import mappingproxy
from collections import Iterator, Iterable

import blaze
from blaze.compute.python import (nunique, mean, rrowfunc, rowfunc,
                                  reduce_by_funcs, optimize)
from blaze import dshape, symbol, discover
from blaze.compatibility import PY2
from blaze.compute.core import compute, compute_up, pre_compute
from blaze.expr import (by, merge, join, distinct, sum, min, max, any, summary,
                        count, std, head, sample, transform, greatest, least)
import numpy as np

from blaze import cos, sin
from blaze.compatibility import builtins
from blaze.utils import raises


t = symbol('t', 'var * {name: string, amount: int, id: int}')


data = [['Alice', 100, 1],
        ['Bob', 200, 2],
        ['Alice', 50, 3]]


tbig = symbol('tbig', 'var * {name: string, sex: string[1], amount: int, id: int}')


databig = [['Alice', 'F', 100, 1],
           ['Alice', 'F', 100, 3],
           ['Drew', 'F', 100, 4],
           ['Drew', 'M', 100, 5],
           ['Drew', 'M', 200, 5]]


def test_dispatched_rowfunc():
    cw = optimize(t['amount'] + 100, [])
    assert rowfunc(t)(t) == t
    assert rowfunc(cw)(('Alice', 100, 1)) == 200


def test_reduce_by_funcs():
    e = summary(number=t.id.max(), sum=t.amount.sum())
    b = by(t, e)
    assert reduce_by_funcs(b)[2]([1,2,3], [4,5,6]) == (1, 7)


def test_symbol():
    assert compute(t, data) == data


def test_projection():
    assert list(compute(t['name'], data)) == [x[0] for x in data]


def test_eq():
    assert list(compute(t['amount'] == 100, data)) == [x[1] == 100 for x in data]


def test_selection():
    assert list(compute(t[t['amount'] == 0], data)) == \
                [x for x in data if x[1] == 0]
    assert list(compute(t[t['amount'] > 150], data)) == \
                [x for x in data if x[1] > 150]


def test_arithmetic():
    assert list(compute(t['amount'] + t['id'], data)) == \
                [b + c for a, b, c, in data]
    assert list(compute(t['amount'] * t['id'], data)) == \
                [b * c for a, b, c, in data]
    assert list(compute(t['amount'] % t['id'], data)) == \
                [b % c for a, b, c, in data]


def test_unary_ops():
    for op in ('cos', 'sin', 'exp', 'ceil', 'floor', 'trunc', 'isnan'):
        f = getattr(blaze, op)
        pyf = getattr(math, op)
        result = list(compute(f(t['amount']), data))
        assert result == [pyf(x[1]) for x in data]


def test_neg():
    expr = optimize(-t.amount, [])
    assert list(compute(expr, data)) == [-x[1] for x in data]


def test_reductions():
    assert compute(sum(t['amount']), data) == 100 + 200 + 50
    assert compute(min(t['amount']), data) == 50
    assert compute(max(t['amount']), data) == 200
    assert compute(nunique(t['amount']), data) == 3
    assert compute(nunique(t['name']), data) == 2
    assert compute(count(t['amount']), data) == 3
    assert compute(any(t['amount'] > 150), data) is True
    assert compute(any(t['amount'] > 250), data) is False
    assert compute(t.amount[0], data) == 100
    assert compute(t.amount[-1], data) == 50


def test_1d_reductions_keepdims():
    for r in [sum, min, max, nunique, count]:
        assert compute(r(t.amount, keepdims=True), data) == \
               (compute(r(t.amount), data),)


def test_count():
    t = symbol('t', '3 * int')
    assert compute(t.count(), [1, None, 2]) == 2


def reduction_runner(funcs):
    from blaze.compatibility import builtins as bts
    exprs = sum, min, max
    for blaze_expr, py_func in itertools.product(exprs, funcs):
        f = getattr(operator, py_func)
        reduc_f = getattr(bts, blaze_expr.__name__)
        ground_truth = f(reduc_f([100, 200, 50]), 5)
        assert compute(f(blaze_expr(t['amount']), 5), data) == ground_truth


def test_reduction_arithmetic():
    funcs = 'add', 'mul'
    reduction_runner(funcs)


def test_reduction_compare():
    funcs = 'eq', 'ne', 'lt', 'gt', 'le', 'ge'
    reduction_runner(funcs)


def test_mean():
    assert compute(mean(t['amount']), data) == float(100 + 200 + 50) / 3
    assert 50 < compute(std(t['amount']), data) < 100


def test_std():
    amt = [row[1] for row in data]
    assert np.allclose(compute(t.amount.std(), data), np.std(amt))
    assert np.allclose(compute(t.amount.std(unbiased=True), data),
                       np.std(amt, ddof=1))
    assert np.allclose(compute(t.amount.var(), data), np.var(amt))
    assert np.allclose(compute(t.amount.var(unbiased=True), data),
                       np.var(amt, ddof=1))


def test_by_no_grouper():
    names = t['name']
    assert set(compute(by(names, count=names.count()), data)) == \
            set([('Alice', 2), ('Bob', 1)])


def test_by_one():
    print(compute(by(t['name'], total=t['amount'].sum()), data))
    assert set(compute(by(t['name'], total=t['amount'].sum()), data)) == \
            set([('Alice', 150), ('Bob', 200)])


def test_by_compound_apply():
    print(compute(by(t['name'], total=(t['amount'] + 1).sum()), data))
    assert set(compute(by(t['name'], total=(t['amount'] + 1).sum()), data)) == \
            set([('Alice', 152), ('Bob', 201)])


def test_by_two():
    result = compute(by(tbig[['name', 'sex']], total=tbig['amount'].sum()),
                     databig)

    expected = [('Alice', 'F', 200),
                ('Drew', 'F', 100),
                ('Drew', 'M', 300)]

    print(set(result))
    assert set(result) == set(expected)


def test_by_three():
    result = compute(by(tbig[['name', 'sex']],
                        total=(tbig['id'] + tbig['amount']).sum()),
                     databig)

    expected = [('Alice', 'F', 204),
                ('Drew', 'F', 104),
                ('Drew', 'M', 310)]

    print(result)
    assert set(result) == set(expected)


def test_works_on_generators():
    assert list(compute(t['amount'], iter(data))) == \
            [x[1] for x in data]
    assert list(compute(t['amount'], (i for i in data))) == \
            [x[1] for x in data]


def test_join():
    left = [['Alice', 100], ['Bob', 200]]
    right = [['Alice', 1], ['Bob', 2]]

    L = symbol('L', 'var * {name: string, amount: int}')
    R = symbol('R', 'var * {name: string, id: int}')
    joined = join(L, R, 'name')

    assert dshape(joined.schema) == \
            dshape('{name: string, amount: int, id: int}')

    result = list(compute(joined, {L: left, R: right}))

    expected = [('Alice', 100, 1), ('Bob', 200, 2)]

    assert result == expected


def test_outer_join():
    left = [(1, 'Alice', 100),
            (2, 'Bob', 200),
            (4, 'Dennis', 400)]
    right = [('NYC', 1),
             ('Boston', 1),
             ('LA', 3),
             ('Moscow', 4)]

    L = symbol('L', 'var * {id: int, name: string, amount: real}')
    R = symbol('R', 'var * {city: string, id: int}')

    assert set(compute(join(L, R), {L: left, R: right})) == set(
            [(1, 'Alice', 100, 'NYC'),
             (1, 'Alice', 100, 'Boston'),
             (4, 'Dennis', 400, 'Moscow')])

    assert set(compute(join(L, R, how='left'), {L: left, R: right})) == set(
            [(1, 'Alice', 100, 'NYC'),
             (1, 'Alice', 100, 'Boston'),
             (2, 'Bob', 200, None),
             (4, 'Dennis', 400, 'Moscow')])

    assert set(compute(join(L, R, how='right'), {L: left, R: right})) == set(
            [(1, 'Alice', 100, 'NYC'),
             (1, 'Alice', 100, 'Boston'),
             (3, None, None, 'LA'),
             (4, 'Dennis', 400, 'Moscow')])

    assert set(compute(join(L, R, how='outer'), {L: left, R: right})) == set(
            [(1, 'Alice', 100, 'NYC'),
             (1, 'Alice', 100, 'Boston'),
             (2, 'Bob', 200, None),
             (3, None, None, 'LA'),
             (4, 'Dennis', 400, 'Moscow')])


def test_multi_column_join():
    left = [(1, 2, 3),
            (2, 3, 4),
            (1, 3, 5)]
    right = [(1, 2, 30),
             (1, 3, 50),
             (1, 3, 150)]

    L = symbol('L', 'var * {x: int, y: int, z: int}')
    R = symbol('R', 'var * {x: int, y: int, w: int}')

    j = join(L, R, ['x', 'y'])

    print(list(compute(j, {L: left, R: right})))
    assert list(compute(j, {L: left, R: right})) == [(1, 2, 3, 30),
                                                     (1, 3, 5, 50),
                                                     (1, 3, 5, 150)]


@pytest.mark.xfail(reason="This doesn't necessarily make sense")
def test_column_of_column():
    assert list(compute(t['name']['name'], data)) == \
            list(compute(t['name'], data))


def test_distinct():
    assert set(compute(distinct(t['name']), data)) == set(['Alice', 'Bob'])
    assert set(compute(distinct(t), data)) == set(map(tuple, data))
    e = distinct(t)
    assert list(compute(e, [])) == []


def test_distinct_count():
    t2 = t['name'].distinct()
    gby = by(t2, total=t2.count())
    result = set(compute(gby, data))
    assert result == set([('Alice', 1), ('Bob', 1)])


def test_sort():
    assert list(compute(t.sort('amount'), data)) == \
            sorted(data, key=lambda x: x[1], reverse=False)

    assert list(compute(t.sort('amount', ascending=True), data)) == \
            sorted(data, key=lambda x: x[1], reverse=False)

    assert list(compute(t.sort(['amount', 'id']), data)) == \
            sorted(data, key=lambda x: (x[1], x[2]), reverse=False)


def test_fancy_sort():
    assert list(compute(t.sort(t['amount']), data)) ==\
            list(compute(t.sort('amount'), data))

    assert list(compute(t.sort(t[['amount', 'id']]), data)) ==\
            list(compute(t.sort(['amount', 'id']), data))

    assert list(compute(t.sort(0-t['amount']), data)) ==\
            list(compute(t.sort('amount'), data))[::-1]


def test_sort_on_column():
    assert list(compute(t.name.distinct().sort('name'), data)) == \
            ['Alice', 'Bob']


def test_head():
    assert list(compute(t.head(1), data)) == [data[0]]

    e = head(t, 101)
    p = list(range(1000))
    assert len(list(compute(e, p))) == 101


def test_sample():
    NN = len(databig)
    for n in range(1, NN+1):
        assert (len(compute(tbig.sample(n=n), databig)) ==
                len(compute(tbig.sample(frac=float(n)/NN), databig)) ==
                n)
    assert len(compute(tbig.sample(n=NN*2), databig)) == NN


def test_graph_double_join():
    idx = [['A', 1],
           ['B', 2],
           ['C', 3],
           ['D', 4],
           ['E', 5],
           ['F', 6]]

    arc = [[1, 3],
           [2, 3],
           [4, 3],
           [5, 3],
           [3, 1],
           [2, 1],
           [5, 1],
           [1, 6],
           [2, 6],
           [4, 6]]

    wanted = [['A'],
              ['F']]

    t_idx = symbol('t_idx', 'var * {name: string, b: int32}')
    t_arc = symbol('t_arc', 'var * {a: int32, b: int32}')
    t_wanted = symbol('t_wanted', 'var * {name: string}')

    # >>> compute(join(t_idx, t_arc, 'b'), {t_idx: idx, t_arc: arc})
    # [[1, A, 3],
    #  [1, A, 2],
    #  [1, A, 5],
    #  [3, C, 1],
    #  [3, C, 2],
    #  [3, C, 4],
    #  [3, C, 5],
    #  [6, F, 1],
    #  [6, F, 2],
    #  [6, F, 4]]

    j = join(join(t_idx, t_arc, 'b'), t_wanted, 'name')[['name', 'b', 'a']]

    result = compute(j, {t_idx: idx, t_arc: arc, t_wanted: wanted})
    result = sorted(map(tuple, result))
    expected = sorted([('A', 1, 3),
                       ('A', 1, 2),
                       ('A', 1, 5),
                       ('F', 6, 1),
                       ('F', 6, 2),
                       ('F', 6, 4)])

    assert result == expected


def test_label():
    assert list(compute((t['amount'] * 1).label('foo'), data)) == \
            list(compute((t['amount'] * 1), data))


def test_relabel_join():
    names = symbol('names', 'var * {first: string, last: string}')

    siblings = join(names.relabel({'first': 'left'}),
                    names.relabel({'first': 'right'}),
                    'last')[['left', 'right']]

    data = [('Alice', 'Smith'),
            ('Bob', 'Jones'),
            ('Charlie', 'Smith')]

    print(set(compute(siblings, {names: data})))
    assert ('Alice', 'Charlie') in set(compute(siblings, {names: data}))
    assert ('Alice', 'Bob') not in set(compute(siblings, {names: data}))


def test_map_column():
    inc = lambda x: x + 1
    assert list(compute(t['amount'].map(inc, 'int'), data)) == [x[1] + 1 for x in data]


def test_map():
    assert (list(compute(t.map(lambda tup: tup[1] + tup[2], 'int'), data)) ==
            [x[1] + x[2] for x in data])


def test_apply_column():
    result = compute(t.amount.apply(builtins.sum, 'real'), data)
    expected = compute(t.amount.sum(), data)

    assert result == expected


def test_apply():
    data2 = tuple(map(tuple, data))
    assert compute(t.apply(hash, 'int'), data2) == hash(data2)


def test_map_datetime():
    from datetime import datetime
    data = [['A', 0], ['B', 1]]
    t = symbol('t', 'var * {foo: string, datetime: int64}')

    result = list(compute(t['datetime'].map(datetime.utcfromtimestamp,
    'datetime'), data))
    expected = [datetime(1970, 1, 1, 0, 0, 0), datetime(1970, 1, 1, 0, 0, 1)]

    assert result == expected


def test_by_multi_column_grouper():
    t = symbol('t', 'var * {x: int, y: int, z: int}')
    expr = by(t[['x', 'y']], total=t['z'].count())
    data = [(1, 2, 0), (1, 2, 0), (1, 1, 0)]

    print(set(compute(expr, data)))
    assert set(compute(expr, data)) == set([(1, 2, 2), (1, 1, 1)])


def test_merge():
    col = (t['amount'] * 2).label('new')

    expr = merge(t['name'], col)

    assert list(compute(expr, data)) == [(row[0], row[1] * 2) for row in data]


def test_transform():
    expr = transform(t, x=t.amount / t.id)
    assert list(compute(expr, data)) == [('Alice', 100, 1, 100),
                                         ('Bob',   200, 2, 100),
                                         ('Alice',  50, 3, 50 / 3)]


def test_map_columnwise():
    colwise = t['amount'] * t['id']

    expr = colwise.map(lambda x: x / 10, 'int64', name='mod')

    assert list(compute(expr, data)) == [((row[1]*row[2]) / 10) for row in data]


def test_map_columnwise_of_selection():
    tsel = t[t['name'] == 'Alice']
    colwise = tsel['amount'] * tsel['id']

    expr = colwise.map(lambda x: x / 10, 'int64', name='mod')

    assert list(compute(expr, data)) == [((row[1]*row[2]) / 10) for row in data[::2]]


def test_selection_out_of_order():
    expr = t['name'][t['amount'] < 100]

    assert list(compute(expr, data)) == ['Alice']


def test_recursive_rowfunc():
    f = rrowfunc(t['name'], t)
    assert [f(row) for row in data] == [row[0] for row in data]

    expr = optimize(t['amount'] + t['id'], [])
    f = rrowfunc(expr, t)
    assert [f(row) for row in data] == [row[1] + row[2] for row in data]

    assert raises(Exception, lambda: rrowfunc(t[t['amount'] < 0]['name'], t))


def test_recursive_rowfunc_is_used():
    expr = by(t['name'], total=(2 * (t['amount'] + t['id'])).sum())
    expected = [('Alice', 2*(101 + 53)),
                ('Bob', 2*(202))]
    assert set(compute(expr, data)) == set(expected)


class TestFunctionExpressions(object):
    def test_compound(self):
        s = t.amount.mean()
        r = compute(s, data)
        assert isinstance(r, float)

        expr = cos(s) ** 2 + sin(s) ** 2
        result = compute(expr, data)
        expected = math.cos(r) ** 2 + math.sin(r) ** 2
        assert result == expected

    def test_user_defined_function(self):
        s = t.amount.count()
        r = compute(s, data)
        assert isinstance(r, int)

        def myfunc(x):
            return (cos(x) + sin(x)) ** 2 / math.pi

        result = compute(myfunc(s), data)
        expected = (math.cos(r) + math.sin(r)) ** 2 / math.pi
        assert result == expected

    def test_user_defined_calls(self):
        s = t.amount.count()
        r = compute(s, data)

        def myother(y):
            return 2 + y ** 10

        def myfunc(x):
            return myother((cos(x) + sin(x)) ** 2 / math.pi)

        result = compute(myfunc(s), data)
        expected = myother((math.cos(r) + math.sin(r)) ** 2 / math.pi)
        assert result == expected


def test_by_groupby_deep():
    data = [(1, 2, 'Alice'),
            (1, 3, 'Bob'),
            (2, 4, 'Alice'),
            (2, 4, '')]

    schema = '{x: int, y: int, name: string}'
    t = symbol('t', datashape.var * schema)

    t2 = t[t['name'] != '']
    t3 = merge(t2.x, t2.name)
    expr = by(t3.name, avg=t3.x.mean())
    result = set(compute(expr, data))
    assert result == set([('Alice', 1.5), ('Bob', 1.0)])


def test_by_then_sort_dict_items_sequence():
    expr = by(tbig.name, total=tbig.amount.sum()).sort('name')
    assert compute(expr, databig)


def test_summary():
    expr = summary(count=t.id.count(), sum=t.amount.sum())
    assert compute(expr, data) == (3, 350)
    assert compute(expr, iter(data)) == (3, 350)


def test_summary_keepdims():
    assert compute(summary(count=t.id.count(), sum=t.amount.sum(),
                           keepdims=True), data) == \
          (compute(summary(count=t.id.count(), sum=t.amount.sum(),
                           keepdims=False), data),)


def test_summary_by():
    expr = by(t.name, summary(count=t.id.count(), sum=t.amount.sum()))
    assert set(compute(expr, data)) == set([('Alice', 2, 150),
                                            ('Bob', 1, 200)])

    expr = by(t.name, summary(count=t.id.count(), sum=(t.amount + 1).sum()))
    assert set(compute(expr, data)) == set([('Alice', 2, 152),
                                            ('Bob', 1, 201)])

    expr = by(t.name, summary(count=t.id.count(), sum=t.amount.sum() + 1))
    assert set(compute(expr, data)) == set([('Alice', 2, 151),
                                            ('Bob', 1, 201)])


def test_summary_by_first():
    expr = by(t.name, amt=t.amount[0])
    assert set(compute(expr, data)) == set((('Bob', 200), ('Alice', 100)))


def test_summary_by_last():
    expr = by(t.name, amt=t.amount[-1])
    assert set(compute(expr, data)) == set((('Bob', 200), ('Alice', 50)))


def test_reduction_arithmetic():
    expr = t.amount.sum() + 1
    assert compute(expr, data) == 351


def test_scalar_arithmetic():
    x = symbol('x', 'real')
    y = symbol('y', 'real')
    assert compute(x + y, {x: 2, y: 3}) == 5
    assert compute_up(x + y, 2, 3) == 5
    assert compute_up(x * y, 2, 3) == 6
    assert compute_up(x / y, 6, 3) == 2
    assert compute_up(x % y, 4, 3) == 1
    assert compute_up(x ** y, 4, 3) == 64

    assert compute(x + 1, {x: 2}) == 3
    assert compute(x * 2, {x: 2}) == 4
    assert compute(1 + x, {x: 2}) == 3
    assert compute(2 * x, {x: 2}) == 4

    assert compute_up(-x, 1) == -1

    assert compute_up(blaze.sin(x), 1) == math.sin(1)


def test_like():
    t = symbol('t', 'var * {name: string, city: string}')
    data = [
        ('Alice Smith', 'New York'),
        ('Bob Smith', 'Chicago'),
        ('Alice Walker', 'LA')
    ]

    assert list(compute(t[t.name.like('Alice*')], data)) == [data[0], data[2]]
    assert list(compute(t[t.name.like('lice*')], data)) == []
    assert list(compute(t[t.name.like('*Smith*')], data)) == [data[0], data[1]]
    assert list(
        compute(t[t.name.like('*Smith*') & t.city.like('New York')], data)
    ) == [data[0]]


def test_datetime_comparison():
    data = [['Alice', date(2000, 1, 1)],
            ['Bob', date(2000, 2, 2)],
            ['Alice', date(2000, 3, 3)]]

    t = symbol('t', 'var * {name: string, when: date}')

    assert list(compute(t[t.when > '2000-01-01'], data)) == data[1:]


def test_datetime_access():
    data = [['Alice', 100, 1, datetime(2000, 1, 1, 1, 1, 1)],
            ['Bob', 200, 2, datetime(2000, 1, 1, 1, 1, 1)],
            ['Alice', 50, 3, datetime(2000, 1, 1, 1, 1, 1)]]

    t = symbol('t',
            'var * {amount: float64, id: int64, name: string, when: datetime}')

    assert list(compute(t.when.year, data)) == [2000, 2000, 2000]
    assert list(compute(t.when.second, data)) == [1, 1, 1]
    assert list(compute(t.when.date, data)) == [date(2000, 1, 1)] * 3


def test_utcfromtimestamp():
    t = symbol('t', '1 * int64')
    assert list(compute(t.utcfromtimestamp, [0])) == \
            [datetime(1970, 1, 1, 0, 0)]


payments = [{'name': 'Alice', 'payments': [
                {'amount':  100, 'when': datetime(2000, 1, 1, 1, 1 ,1)},
                {'amount':  200, 'when': datetime(2000, 2, 2, 2, 2, 2)}
                ]},
            {'name': 'Bob', 'payments': [
                {'amount':  300, 'when': datetime(2000, 3, 3, 3, 3 ,3)},
                {'amount': -400, 'when': datetime(2000, 4, 4, 4, 4, 4)},
                {'amount':  500, 'when': datetime(2000, 5, 5, 5, 5, 5)}
                ]},
            ]

payments_ordered = [('Alice', [( 100, datetime(2000, 1, 1, 1, 1 ,1)),
                               ( 200, datetime(2000, 2, 2, 2, 2, 2))]),
                    ('Bob',   [( 300, datetime(2000, 3, 3, 3, 3 ,3)),
                               (-400, datetime(2000, 4, 4, 4, 4, 4)),
                               ( 500, datetime(2000, 5, 5, 5, 5, 5))])]

payment_dshape = 'var * {name: string, payments: var * {amount: int32, when: datetime}}'


@pytest.mark.xfail(reason="Can't reason about nested broadcasts yet")
def test_nested():
    t = symbol('t', payment_dshape)
    assert list(compute(t.name, payments_ordered)) == ['Alice', 'Bob']

    assert list(compute(t.payments, payments_ordered)) == \
                [p[1] for p in payments_ordered]
    assert list(compute(t.payments.amount, payments_ordered)) == \
            [(100, 200), (300, -400, 500)]
    assert list(compute(t.payments.amount + 1, payments_ordered)) ==\
            [(101, 201), (301, -399, 501)]


@pytest.mark.xfail(reason="Can't reason about nested broadcasts yet")
def test_scalar():
    s = symbol('s', '{name: string, id: int32, payments: var * {amount: int32, when: datetime}}')
    data = ('Alice', 1, ((100, datetime(2000, 1, 1, 1, 1 ,1)),
                         (200, datetime(2000, 2, 2, 2, 2, 2)),
                         (300, datetime(2000, 3, 3, 3, 3, 3))))

    assert compute(s.name, data) == 'Alice'
    assert compute(s.id + 1, data) == 2
    assert tuple(compute(s.payments.amount, data)) == (100, 200, 300)
    assert tuple(compute(s.payments.amount + 1, data)) == (101, 201, 301)


def test_slice():
    assert compute(t[0], data) == data[0]
    assert list(compute(t[:2], data)) == list(data[:2])
    assert list(compute(t.name[:2], data)) == [data[0][0], data[1][0]]


def test_negative_slicing():
    assert list(compute(t[-1:], data)) == data[-1:]
    assert list(compute(t[-1:], iter(data))) == data[-1:]
    assert list(compute(t[-1], data)) == data[-1]
    assert list(compute(t[-1], iter(data))) == data[-1]
    assert list(compute(t[-2], data)) == data[-2]
    assert list(compute(t[-2], iter(data))) == data[-2]


@pytest.mark.xfail(raises=ValueError,
                   reason="No support for stop and step having negative values")
def test_negative_slicing_raises_on_stop_and_step_not_None():
    assert list(compute(t[-2:-5:-1], data)) == data[-2:-5:-1]


def test_multi_dataset_broadcast():
    x = symbol('x', '3 * int')
    y = symbol('y', '3 * int')

    a = [1, 2, 3]
    b = [10, 20, 30]

    assert list(compute(x + y, {x: a, y: b})) == [11, 22, 33]
    assert list(compute(2*x + (y + 1), {x: a, y: b})) == [13, 25, 37]


@pytest.mark.xfail(reason="Optimize doesn't create multi-table-broadcasts")
def test_multi_dataset_broadcast_with_Record_types():
    x = symbol('x', '3 * {p: int, q: int}')
    y = symbol('y', '3 * int')

    a = [(1, 1), (2, 2), (3, 3)]
    b = [10, 20, 30]

    assert list(compute(x.p + x.q + y, {x: iter(a), y: iter(b)})) == [12, 24, 36]


def eq(a, b):
    if isinstance(a, (Iterable, Iterator)):
        a = list(a)
    if isinstance(b, (Iterable, Iterator)):
        b = list(b)
    return a == b


def test_pre_compute():
    s = symbol('s', 'var * {a: int, b: int}')
    assert pre_compute(s, [(1, 2)]) == [(1, 2)]
    assert list(pre_compute(s, iter([(1, 2)]))) == [(1, 2)]
    assert list(pre_compute(s, iter([(1, 2), (3, 4)]))) == [(1, 2), (3, 4)]
    assert list(pre_compute(s, iter([{'a': 1, 'b': 2},
                                     {'a': 3, 'b': 4}]))) == [(1, 2), (3, 4)]


def test_dicts():
    t = symbol('t', 'var * {name: string, amount: int, id: int}')

    L = [['Alice', 100, 1],
         ['Bob', 200, 2],
         ['Alice', 50, 3]]

    d = [{'name': 'Alice', 'amount': 100, 'id': 1},
         {'name': 'Bob', 'amount': 200, 'id': 2},
         {'name': 'Alice', 'amount': 50, 'id': 3}]

    assert list(pre_compute(t, d)) == list(map(tuple, L))

    for expr in [t.amount, t.amount.sum(), by(t.name, sum=t.amount.sum())]:
        assert eq(compute(expr, {t: L}),
                  compute(expr, {t: d}))

    for expr in [t.amount, t.amount.sum(), by(t.name, sum=t.amount.sum())]:
        assert eq(compute(expr, {t: iter(L)}),
                  compute(expr, {t: iter(d)}))
        assert eq(compute(expr, {t: iter(L)}),
                  compute(expr, {t: L}))


def test_nelements_list_tuple():
    assert compute(t.nelements(), data) == len(data)


def test_nelements_iterator():
    x = (row for row in data)
    assert compute(t.nelements(), x) == len(data)


def test_nrows():
    assert compute(t.nrows, data) == len(data)
    x = (row for row in data)
    assert compute(t.nrows, x) == len(data)


@pytest.mark.xfail(raises=Exception, reason="Only 1D reductions allowed")
def test_nelements_2D():
    assert compute(t.nelements(axis=1), data) == len(data[0])


def test_compute_field_on_dicts():
    s = symbol('s', '{x: 3 * int, y: 3 * int}')
    d = {'x': [1, 2, 3], 'y': [4, 5, 6]}
    assert compute(s.x, {s: d}) == [1, 2, 3]


class _MapProxy(Mapping):  # pragma: no cover
    def __init__(self, mapping):
        self._map = mapping

    def __getitem__(self, key):
        return self._map[key]

    def __iter__(self):
        return iter(self._map)

    def __len__(self):
        return len(self._map)


@pytest.mark.parametrize(
    'cls',
    # mappingproxy refers to PyDictProxy in py2 which is not registered
    # as a mapping
    (_MapProxy,) + (mappingproxy,) if not PY2 else (),
)
def test_comute_on_mapping(cls):
    s = symbol('s', '{x: 3 * int64, y: 3 * float32}')
    d = cls({'x': [1, 2, 3], 'y': [4.0, 5.0, 6.0]})
    assert compute(s.x, {s: d}, return_type='native') == [1, 2, 3]


def test_truncate():
    s = symbol('x', 'real')
    assert compute(s.truncate(20), 154) == 140
    assert compute(s.truncate(0.1), 3.1415) == 3.1


def test_truncate_datetime():
    s = symbol('x', 'datetime')
    assert compute(s.truncate(2, 'days'), datetime(2002, 1, 3, 12, 30)) ==\
            date(2002, 1, 2)

    s = symbol('x', 'var * datetime')
    assert list(compute(s.truncate(2, 'days'),
                        [datetime(2002, 1, 3, 12, 30)])) ==\
            [date(2002, 1, 2)]


def test_compute_up_on_base():
    d = datetime.now()
    s = symbol('s', 'datetime')
    assert compute(s.minute, d) == d.minute


def test_notnull():
    data = [('Alice', -100, None),
            (None, None, None),
            ('Bob', 300, 'New York City')]
    t = symbol('t', 'var * {name: ?string, amount: ?int32, city: ?string}')
    expr = t.name.notnull()
    result = compute(expr, data)
    assert list(result) == [True, False, True]


def test_notnull_whole_collection():
    t = symbol('t', 'var * {name: ?string, amount: ?int32, city: ?string}')
    with pytest.raises(AttributeError):
        t.notnull


@pytest.mark.parametrize('keys', [['Alice'], ['Bob', 'Alice']])
def test_isin(keys):
    expr = t[t.name.isin(keys)]
    result = list(compute(expr, data))
    expected = [el for el in data if el[0] in keys]
    assert result == expected


def test_greatest():
    s_data, t_data = [1, 2], [2, 1]
    s, t = symbol('s', discover(s_data)), symbol('t', discover(t_data))
    result = compute(greatest(s, t), {s: s_data, t: t_data})
    expected = np.maximum(s_data, t_data).tolist()
    assert list(result) == expected


def test_least():
    s_data, t_data = [1, 2], [2, 1]
    s, t = symbol('s', discover(s_data)), symbol('t', discover(t_data))
    result = compute(least(s, t), {s: s_data, t: t_data})
    expected = np.minimum(s_data, t_data).tolist()
    assert list(result) == expected



from __future__ import absolute_import, division, print_function

from .expressions import ElemWise, schema_method_list, method_properties

import datashape
from datashape import dshape, isdatelike, isnumeric
from datashape.coretypes import timedelta_
from ..compatibility import basestring


__all__ = ['DateTime', 'Date', 'date', 'Year', 'year', 'Month', 'month', 'Day',
           'day', 'days', 'Hour', 'hour', 'Minute', 'minute', 'Second', 'second',
           'Millisecond', 'millisecond', 'Microsecond', 'microsecond', 'nanosecond',
           'Date',
           'date', 'Time', 'time', 'week', 'nanoseconds', 'seconds', 'total_seconds',
           'UTCFromTimestamp', 'DateTimeTruncate',
           'Ceil', 'Floor', 'Round', 'strftime']


def _validate(var, name, type, typename):
    if not isinstance(var, type):
        raise TypeError('"%s" argument must be a %s'%(name, typename))


class DateTime(ElemWise):
    """ Superclass for datetime accessors """
    _arguments = '_child',

    def __str__(self):
        return '%s.%s' % (str(self._child), type(self).__name__.lower())

    def _schema(self):
        ds = dshape(self._dtype)
        return ds if not isinstance(self._child.schema.measure, datashape.Option) else datashape.Option(ds)

    @property
    def _name(self):
        return '%s_%s' % (self._child._name, self.attr)

    @property
    def attr(self):
        return type(self).__name__.lower()


class Date(DateTime):
    _dtype = datashape.date_


def date(expr):
    return Date(expr)


class Year(DateTime):
    _dtype = datashape.int32


def year(expr):
    return Year(expr)


class Month(DateTime):
    _dtype = datashape.int64


def month(expr):
    return Month(expr)


class Day(DateTime):
    _dtype = datashape.int64


def day(expr):
    return Day(expr)


class Time(DateTime):
    _dtype = datashape.time_


def time(expr):
    return Time(expr)


class Hour(DateTime):
    _dtype = datashape.int64


def hour(expr):
    return Hour(expr)


class Minute(DateTime):
    _dtype = datashape.int64


def minute(expr):
    return Minute(expr)


class Second(DateTime):
    _dtype = datashape.int64


def second(expr):
    return Second(expr)


class Millisecond(DateTime):
    _dtype = datashape.int64


def millisecond(expr):
    return Millisecond(expr)


class Microsecond(DateTime):
    _dtype = datashape.int64


def microsecond(expr):
    return Microsecond(expr)


class UTCFromTimestamp(DateTime):
    _dtype = datashape.datetime_


def utcfromtimestamp(expr):
    return UTCFromTimestamp(expr)


class nanosecond(DateTime): _dtype = datashape.int64
class week(DateTime): _dtype = datashape.int64
class weekday(DateTime): _dtype = datashape.int64
class weekday_name(DateTime): _dtype = datashape.string
class daysinmonth(DateTime): _dtype = datashape.int64
class weekofyear(DateTime): _dtype = datashape.int64
class dayofyear(DateTime): _dtype = datashape.int64
class dayofweek(DateTime): _dtype = datashape.int64
class quarter(DateTime): _dtype = datashape.int64
class is_month_start(DateTime): _dtype = datashape.bool_
class is_month_end(DateTime): _dtype = datashape.bool_
class is_quarter_start(DateTime): _dtype = datashape.bool_
class is_quarter_end(DateTime): _dtype = datashape.bool_
class is_year_start(DateTime): _dtype = datashape.bool_
class is_year_end(DateTime): _dtype = datashape.bool_
class days_in_month(DateTime): _dtype = datashape.int64

class strftime(ElemWise):
    _arguments = '_child', 'format'
    schema = datashape.string

units = (
    'year',
    'month',
    'week',
    'day',
    'hour',
    'minute',
    'second',
    'millisecond',
    'microsecond',
    'nanosecond',
)


unit_aliases = {
    'y': 'year',
    'w': 'week',
    'd': 'day',
    'date': 'day',
    'h': 'hour',
    's': 'second',
    'ms': 'millisecond',
    'us': 'microsecond',
    'ns': 'nanosecond'
}


def normalize_time_unit(s):
    """ Normalize time input to one of 'year', 'second', 'millisecond', etc..

    Examples
    --------

    >>> normalize_time_unit('milliseconds')
    'millisecond'
    >>> normalize_time_unit('ms')
    'millisecond'
    """
    s = s.lower().strip()
    if s in units:
        return s
    if s in unit_aliases:
        return unit_aliases[s]
    if s[-1] == 's':
        return normalize_time_unit(s.rstrip('s'))

    raise ValueError("Do not understand time unit %s" % s)


class DateTimeTruncate(DateTime):
    _arguments = '_child', 'measure', 'unit'

    @property
    def _dtype(self):
        if units.index('day') >= units.index(self.unit):
            return datashape.date_
        else:
            return datashape.datetime_

    @property
    def _name(self):
        return self._child._name

    def __str__(self):
        return '%s.truncate(%ss=%g)' % (self._child, self.unit, self.measure)


def truncate(expr, *args, **kwargs):
    """ Truncate datetime expression

    Examples
    --------

    >>> from blaze import symbol, compute
    >>> from datetime import datetime
    >>> s = symbol('s', 'datetime')

    >>> expr = s.truncate(10, 'minutes')
    >>> compute(expr, datetime(2000, 6, 25, 12, 35, 10))
    datetime.datetime(2000, 6, 25, 12, 30)

    >>> expr = s.truncate(1, 'week')
    >>> compute(expr, datetime(2000, 6, 25, 12, 35, 10))
    datetime.date(2000, 6, 25)

    Alternatively use keyword arguments to specify unit and measure

    >>> expr = s.truncate(weeks=2)
    """
    if not args and not kwargs:
        raise TypeError('truncate takes exactly 2 positional arguments, '
                        'e.g., truncate(2, "days") or 1 keyword argument, '
                        'e.g., truncate(days=2)')
    if args:
        if kwargs:
            raise TypeError('Cannot pass both positional and keyword '
                            'arguments; given %s and %s.' % (args, kwargs))
        measure, unit = args
    else:
        [(unit, measure)] = kwargs.items()
    return DateTimeTruncate(expr, measure, normalize_time_unit(unit))


class UnaryDateTimeFunction(ElemWise):

    """DateTime function that only takes a single argument."""
    _arguments = '_child'


class Round(ElemWise):
    _arguments = '_child', 'freq'

    @property
    def schema(self):
        return self._child.schema


class Floor(ElemWise):
    _arguments = '_child', 'freq'

    @property
    def schema(self):
        return self._child.schema


class Ceil(ElemWise):
    _arguments = '_child', 'freq'

    @property
    def schema(self):
        return self._child.schema


class dt_ns(object):

    def __init__(self, field):
        self.field = field

    def year(self):
        return year(self.field)
    def month(self):
        return month(self.field)
    def day(self):
        return day(self.field)
    def hour(self):
        return hour(self.field)
    def minute(self):
        return minute(self.field)
    def date(self):
        return date(self.field)
    def time(self):
        return time(self.field)
    def second(self):
        return second(self.field)
    def millisecond(self):
        return millisecond(self.field)
    def microsecond(self):
        return microsecond(self.field)
    def nanosecond(self):
        return nanosecond(self.field)
    def weekday(self):
        return weekday(self.field)
    def weekday_name(self):
        return weekday_name(self.field)
    def daysinmonth(self):
        return daysinmonth(self.field)
    def weekofyear(self):
        return weekofyear(self.field)
    def dayofyear(self):
        return dayofyear(self.field)
    def dayofweek(self):
        return dayofweek(self.field)
    def quarter(self):
        return quarter(self.field)
    def is_month_start(self):
        return is_month_start(self.field)
    def is_month_end(self):
        return is_month_end(self.field)
    def is_quarter_start(self):
        return is_quarter_start(self.field)
    def is_quarter_end(self):
        return is_quarter_end(self.field)
    def is_year_start(self):
        return is_year_start(self.field)
    def is_year_end(self):
        return is_year_end(self.field)
    def days_in_month(self):
        return days_in_month(self.field)
    def strftime(self, format):
        _validate(format, 'format', basestring, 'string')
        return strftime(self.field, format)
    def truncate(self, *args, **kwargs):
        return truncate(self.field, *args, **kwargs)
    def round(self, freq):
        _validate(freq, 'freq', basestring, 'string')
        return Round(self.field, freq)
    def floor(self, freq):
        _validate(freq, 'freq', basestring, 'string')
        return Floor(self.field, freq)
    def ceil(self, freq):
        _validate(freq, 'freq', basestring, 'string')
        return Ceil(self.field, freq)
    def week(self):
        return week(self.field)

class dt(object):

    __name__ = 'dt'

    def __get__(self, obj, type=None):
        return dt_ns(obj) if obj is not None else self


class days(DateTime): _dtype = datashape.int64
class nanoseconds(DateTime): _dtype = datashape.int64
class seconds(DateTime): _dtype = datashape.int64
class total_seconds(DateTime): _dtype = datashape.int64


class timedelta_ns(object):

    def __init__(self, field):
        self.field = field

    def days(self): return days(self.field)
    def nanoseconds(self): return nanoseconds(self.field)
    def seconds(self): return seconds(self.field)
    def total_seconds(self): return total_seconds(self.field)


class timedelta(object):

    # pandas uses the same 'dt' name for
    # DateTimeProperties and TimedeltaProperties.
    __name__ = 'dt'

    def __get__(self, obj, type=None):
        return timedelta_ns(obj) if obj is not None else self


def isdeltalike(ds):
    return ds == timedelta_

schema_method_list.extend([
    (isdatelike, set([year, month, day, hour, minute, date, time, second,
                      millisecond, microsecond, truncate,
                      dt()])),
    (isnumeric, set([utcfromtimestamp])),
    (isdeltalike, set([timedelta()]))
])

method_properties |= set([year, month, day, hour, minute, second, millisecond,
                          microsecond, date, time, utcfromtimestamp])



import sys
import os
import json

import pytest

import h5py
import numpy as np
import pandas as pd

from datashape import dshape

from blaze import data_spider, discover


@pytest.fixture
def data(tmpdir):
    csvf = tmpdir.join('foo.csv')
    csvf.write('a,b\n1,2\n3,4')
    h5f = tmpdir.join('foo.hdf5')
    data = np.random.randn(10, 2)
    with h5py.File(str(h5f)) as f:
        f.create_dataset(name='fooh5', shape=data.shape,
                         dtype=data.dtype, data=data)
    jsonf = tmpdir.mkdir('sub').join('foo.json')
    jsonf.write(json.dumps([{'a': 2,
                             'b': 3.14,
                             'c': str(pd.Timestamp('now'))},
                            {'a': 2,
                             'b': 4.2,
                             'c': None,
                             'd': 'foobar'}]))
    return tmpdir


@pytest.fixture
def data_with_cycle(data):
    data.join('cycle').mksymlinkto(data)
    return data


def test_data_spider(data):
    result = data_spider(str(data))
    ss = """{
    %r: {
        'foo.csv': var * {a: int64, b: int64},
        'foo.hdf5': {fooh5: 10 * 2 * float64},
        sub: {'foo.json': 2 * {a: int64, b: float64, c: ?datetime, d: ?string}}
    }
}""" % os.path.basename(str(data))
    assert dshape(discover(result)) == dshape(ss)


@pytest.mark.skipif(sys.platform == 'win32',
                    reason='Windows does not have symlinks')
def test_data_spider_cycle(data_with_cycle):
    result = data_spider(str(data_with_cycle), followlinks=True)
    ss = """{
    %r: {
        'foo.csv': var * {a: int64, b: int64},
        'foo.hdf5': {fooh5: 10 * 2 * float64},
        sub: {'foo.json': 2 * {a: int64, b: float64, c: ?datetime, d: ?string}}
    }
}""" % os.path.basename(str(data_with_cycle))
    assert dshape(discover(result)) != dshape(ss)



from __future__ import absolute_import, division, print_function

from datashape import var


class _slice(object):
    """ A hashable slice object

    >>> _slice(0, 10, None)
    0:10
    """
    def __init__(self, start, stop, step):
        self.start = start
        self.stop = stop
        self.step = step

    def __hash__(self):
        return hash((slice, self.start, self.stop, self.step))

    def __str__(self):
        s = ''
        if self.start is not None:
            s = s + str(self.start)
        s = s + ':'
        if self.stop is not None:
            s = s + str(self.stop)
        if self.step is not None:
            s = s + ':' + str(self.step)
        return s

    def __eq__(self, other):
        return (type(self), self.start, self.stop, self.step) == \
               (type(other), other.start, other.stop, other.step)

    def as_slice(self):
        return slice(self.start, self.stop, self.step)

    __repr__ = __str__


class hashable_list(tuple):
    def __str__(self):
        return str(list(self))


def hashable_index(index):
    """ Convert slice-thing into something hashable

    >>> hashable_index(1)
    1

    >>> isinstance(hash(hashable_index((1, slice(10)))), int)
    True
    """
    if type(index) is tuple:  # can't do isinstance due to hashable_list
        return tuple(map(hashable_index, index))
    elif isinstance(index, list):
        return hashable_list(index)
    elif isinstance(index, slice):
        return _slice(index.start, index.stop, index.step)
    return index


def replace_slices(index):
    """
    Takes input from Slice expression and returns either a list,
    slice object, or tuple.

    Examples
    -------
    >>> replace_slices([1, 2, 345, 12])
    [1, 2, 345, 12]
    >>> type(replace_slices(_slice(1, 5, None))) is slice
    True
    >>> type(replace_slices((2, 5))) is tuple
    True

    """
    if isinstance(index, hashable_list):
        return list(index)
    elif isinstance(index, _slice):
        return index.as_slice()
    elif isinstance(index, tuple):
        return tuple(map(replace_slices, index))
    return index


def maxvar(L):
    """

    >>> maxvar([1, 2, var])
    Var()

    >>> maxvar([1, 2, 3])
    3
    """
    if var in L:
        return var
    else:
        return max(L)


def maxshape(shapes):
    """

    >>> maxshape([(10, 1), (1, 10), ()])
    (10, 10)

    >>> maxshape([(4, 5), (5,)])
    (4, 5)
    """
    shapes = [shape for shape in shapes if shape]
    if not shapes:
        return ()
    ndim = max(map(len, shapes))
    shapes = [(1,) * (ndim - len(shape)) + shape for shape in shapes]
    for dims in zip(*shapes):
        if len(set(dims) - set([1])) >= 2:
            raise ValueError("Shapes don't align, %s" % str(dims))
    return tuple(map(maxvar, zip(*shapes)))



from __future__ import absolute_import, division, print_function

import pytest
pytest.importorskip('flask')
pytest.importorskip('flask_cors')

from base64 import b64encode
from copy import copy
# Python 2.7's `io.StringIO` doesn't behave as expected
# with str values
try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO

import datashape
from datashape.util.testing import assert_dshape_equal
import numpy as np
from odo import odo, convert
from datetime import datetime
import pandas as pd
from pandas import DataFrame
from pandas.util.testing import assert_frame_equal
from toolz import pipe, partial

from blaze.dispatch import dispatch
from blaze.compatibility import _inttypes, PY2
from blaze.expr import Expr
from blaze.utils import example
from blaze import discover, symbol, by, CSV, compute, join, into, data
from blaze.server.client import mimetype
from blaze.server.server import Server, to_tree, from_tree, RC
from blaze.server.serialization import all_formats, trusted_formats, fastmsgpack


accounts = DataFrame([['Alice', 100], ['Bob', 200]],
                     columns=['name', 'amount'])

cities = DataFrame([['Alice', 'NYC'], ['Bob', 'LA']],
                   columns=['name', 'city'])

events = DataFrame([[1, datetime(2000, 1, 1, 12, 0, 0)],
                    [2, datetime(2000, 1, 2, 12, 0, 0)]],
                   columns=['value', 'when'])

db = data('sqlite:///' + example('iris.db'))


class DumbResource(object):
    df = DataFrame({'a': np.arange(5),
                    'b': np.arange(5, 10)})

    class NoResource(Exception):
        pass


@convert.register(DataFrame, DumbResource)
def dumb_to_df(d, return_df=None, **kwargs):
    if return_df is None:
        raise DumbResource.NoResource('return_df must be passed')
    to_return = odo(return_df, DataFrame, dshape=discover(d))
    assert_frame_equal(to_return, DumbResource.df)
    return to_return


@dispatch(Expr, DumbResource)
def compute_down(expr, d, **kwargs):
    return dumb_to_df(d, **kwargs)


@discover.register(DumbResource)
def _discover_dumb(d):
    return discover(DumbResource.df)


tdata = {'accounts': accounts,
         'cities': cities,
         'events': events,
         'db': db,
         'dumb': DumbResource()}


@pytest.fixture(scope='module')
def server():
    s = Server(tdata, all_formats)
    s.app.testing = True
    return s


@pytest.fixture(scope='module')
def add_server():
    s = Server(tdata, all_formats, allow_add=True)
    s.app.testing = True
    return s


@pytest.yield_fixture(params=[None, tdata])
def temp_server(request):
    """For when we want to mutate the server"""
    data = request.param
    s = Server(copy(data), formats=all_formats)
    s.app.testing = True
    with s.app.test_client() as c:
        yield c


@pytest.yield_fixture(params=[None, tdata])
def temp_add_server(request):
    """For when we want to mutate the server, and also add datasets to it."""
    data = request.param
    s = Server(copy(data), formats=all_formats, allow_add=True)
    s.app.testing = True
    with s.app.test_client() as c:
        yield c


@pytest.yield_fixture(params=[None, tdata])
def temp_server_with_excfmt(request):
    """With a custom log exception formatter"""
    data = request.param
    log_format_exc = lambda tb: 'CUSTOM TRACEBACK'
    stream = StringIO()
    s = Server(copy(data), formats=all_formats, allow_add=True,
               logfile=stream, log_exception_formatter=log_format_exc)
    s.app.testing = True
    with s.app.test_client() as client:
        yield (client, stream)



@pytest.yield_fixture
def test(server):
    with server.app.test_client() as c:
        yield c


@pytest.yield_fixture
def test_add(add_server):
    with add_server.app.test_client() as c:
        yield c


@pytest.yield_fixture
def iris_server():
    iris = CSV(example('iris.csv'))
    s = Server(iris, all_formats, allow_add=True)
    s.app.testing = True
    with s.app.test_client() as c:
        yield c


def test_datasets(test):
    response = test.get('/datashape')
    assert_dshape_equal(datashape.dshape(response.data.decode('utf-8')),
                        datashape.dshape(discover(tdata)))


@pytest.mark.parametrize('serial', all_formats)
def test_bad_responses(test, serial):

    post = test.post('/compute/accounts.{name}'.format(name=serial.name),
                     data=serial.dumps(500),)
    assert 'OK' not in post.status

    post = test.post('/compute/non-existent-table.{name}'.format(name=serial.name),
                     data=serial.dumps(0))
    assert 'OK' not in post.status

    post = test.post('/compute/accounts.{name}'.format(name=serial.name))
    assert 'OK' not in post.status


def test_to_from_json():
    t = symbol('t', 'var * {name: string, amount: int}')
    assert from_tree(to_tree(t)).isidentical(t)
    assert from_tree(to_tree(t.amount + 1)).isidentical(t.amount + 1)


def test_to_tree():
    t = symbol('t', 'var * {name: string, amount: int32}')
    expr = t.amount.sum()
    dshape = datashape.dshape('var * {name: string, amount: int32}')
    sum_args = [{'op': 'Field',
                 'args': [{'op': 'Symbol',
                           'args': ['t', dshape, 0]},
                          'amount']},
                [0],
                False]
    expected = {'op': 'sum', 'args': sum_args}
    assert to_tree(expr) == expected


@pytest.mark.parametrize('serial', all_formats)
def test_to_tree_slice(serial):
    t = symbol('t', 'var * {name: string, amount: int32}')
    expr = t[:5]
    expr2 = pipe(expr,
                 partial(to_tree, names={t: 't'}),
                 serial.dumps,
                 serial.loads,
                 partial(from_tree, namespace={'t': t}))
    assert expr.isidentical(expr2)


def test_to_from_tree_namespace():
    t = symbol('t', 'var * {name: string, amount: int32}')
    expr = t.name

    tree = to_tree(expr, names={t: 't'})
    assert tree == {'op': 'Field', 'args': ['t', 'name']}

    new = from_tree(tree, namespace={'t': t})
    assert new.isidentical(expr)

t = symbol('t', discover(tdata))


@pytest.mark.parametrize('serial', all_formats)
def test_compute(test, serial):
    expr = t.accounts.amount.sum()
    query = {'expr': to_tree(expr)}
    expected = 300

    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))

    assert 'OK' in response.status
    tdata = serial.loads(response.data)
    assert serial.data_loads(tdata['data']) == expected
    assert list(tdata['names']) == ['amount_sum']


@pytest.mark.parametrize('serial', all_formats)
def test_get_datetimes(test, serial):
    expr = t.events
    query = {'expr': to_tree(expr)}

    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))

    assert 'OK' in response.status
    tdata = serial.loads(response.data)
    ds = datashape.dshape(tdata['datashape'])
    result = into(np.ndarray,
                  serial.data_loads(tdata['data']),
                  dshape=ds)
    assert into(list, result) == into(list, events)
    assert list(tdata['names']) == events.columns.tolist()


@pytest.mark.parametrize('serial', all_formats)
def dont_test_compute_with_namespace(test, serial):
    query = {'expr': {'op': 'Field',
                      'args': ['accounts', 'name']}}
    expected = ['Alice', 'Bob']

    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))

    assert 'OK' in response.status
    tdata = serial.loads(response.data)
    assert serial.data_loads(tdata['data']) == expected
    assert tdata['names'] == ['name']


iris = CSV(example('iris.csv'))


@pytest.mark.parametrize('serial', all_formats)
def test_compute_with_variable_in_namespace(iris_server, serial):
    test = iris_server
    t = symbol('t', discover(iris))
    pl = symbol('pl', 'float32')
    expr = t[t.petal_length > pl].species
    tree = to_tree(expr, {pl: 'pl'})

    blob = serial.dumps({'expr': tree, 'namespace': {'pl': 5}})
    resp = test.post('/compute',
                     data=blob,
                     headers=mimetype(serial))

    assert 'OK' in resp.status
    tdata = serial.loads(resp.data)
    result = serial.data_loads(tdata['data'])
    expected = list(compute(expr._subs({pl: 5}), {t: iris}))
    assert odo(result, list) == expected
    assert list(tdata['names']) == ['species']


@pytest.mark.parametrize('serial', all_formats)
def test_compute_by_with_summary(iris_server, serial):
    test = iris_server
    t = symbol('t', discover(iris))
    expr = by(t.species,
              max=t.petal_length.max(),
              sum=t.petal_width.sum())
    tree = to_tree(expr)
    blob = serial.dumps({'expr': tree})
    resp = test.post('/compute',
                     data=blob,
                     headers=mimetype(serial))
    assert 'OK' in resp.status
    tdata = serial.loads(resp.data)
    result = DataFrame(serial.data_loads(tdata['data'])).values
    expected = compute(expr, iris).values
    np.testing.assert_array_equal(result[:, 0],
                                  expected[:, 0])
    np.testing.assert_array_almost_equal(result[:, 1:],
                                         expected[:, 1:])
    assert list(tdata['names']) == ['species', 'max', 'sum']


@pytest.mark.parametrize('serial', all_formats)
def test_compute_column_wise(iris_server, serial):
    test = iris_server
    t = symbol('t', discover(iris))
    subexpr = ((t.petal_width / 2 > 0.5) &
               (t.petal_length / 2 > 0.5))
    expr = t[subexpr]
    tree = to_tree(expr)
    blob = serial.dumps({'expr': tree})
    resp = test.post('/compute',
                     data=blob,
                     headers=mimetype(serial))

    assert 'OK' in resp.status
    tdata = serial.loads(resp.data)
    result = serial.data_loads(tdata['data'])
    expected = compute(expr, iris)
    assert list(map(tuple, into(list, result))) == into(list, expected)
    assert list(tdata['names']) == t.fields


@pytest.mark.parametrize('serial', all_formats)
def test_multi_expression_compute(test, serial):
    s = symbol('s', discover(tdata))

    expr = join(s.accounts, s.cities)

    resp = test.post('/compute',
                     data=serial.dumps({'expr': to_tree(expr)}),
                     headers=mimetype(serial))

    assert 'OK' in resp.status
    respdata = serial.loads(resp.data)
    result = serial.data_loads(respdata['data'])
    expected = compute(expr, {s: tdata})

    assert list(map(tuple, odo(result, list))) == into(list, expected)
    assert list(respdata['names']) == expr.fields


@pytest.mark.parametrize('serial', all_formats)
def test_leaf_symbol(test, serial):
    query = {'expr': {'op': 'Field', 'args': [':leaf', 'cities']}}
    resp = test.post('/compute',
                     data=serial.dumps(query),
                     headers=mimetype(serial))

    tdata = serial.loads(resp.data)
    a = serial.data_loads(tdata['data'])
    b = into(list, cities)

    assert list(map(tuple, into(list, a))) == b
    assert list(tdata['names']) == cities.columns.tolist()


@pytest.mark.parametrize('serial', all_formats)
def test_sqlalchemy_result(test, serial):
    expr = t.db.iris.head(5)
    query = {'expr': to_tree(expr)}

    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))

    assert 'OK' in response.status
    tdata = serial.loads(response.data)
    result = serial.data_loads(tdata['data'])
    if isinstance(result, list):
        assert all(isinstance(item, (tuple, list)) for item in result)
    elif isinstance(result, DataFrame):
        expected = DataFrame([[5.1, 3.5, 1.4, 0.2, 'Iris-setosa'],
                              [4.9, 3.0, 1.4, 0.2, 'Iris-setosa'],
                              [4.7, 3.2, 1.3, 0.2, 'Iris-setosa'],
                              [4.6, 3.1, 1.5, 0.2, 'Iris-setosa'],
                              [5.0, 3.6, 1.4, 0.2, 'Iris-setosa']],
                             columns=['sepal_length',
                                      'sepal_width',
                                      'petal_length',
                                      'petal_width',
                                      'species'])
        assert_frame_equal(expected, result)
    assert list(tdata['names']) == t.db.iris.fields


def test_server_accepts_non_nonzero_ables():
    Server(DataFrame())


def serialize_query_with_map_builtin_function(test, serial, fcn):
    """
    serialize a query that invokes the 'map' operation using a builtin function
    return the result of the post operation along with expected result
    """
    t = symbol('t', discover(iris))
    expr = t.species.map(fcn, 'int')
    query = {'expr': to_tree(expr)}
    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))
    assert 'OK' in response.status
    respdata = serial.loads(response.data)
    result = serial.data_loads(respdata['data'])

    exp_res = compute(expr, {t: iris}, return_type=list)
    return (exp_res, result)


@pytest.mark.parametrize('serial', trusted_formats)
def test_map_builtin_client_server(iris_server, serial):
    exp_res, result = serialize_query_with_map_builtin_function(iris_server,
                                                                serial,
                                                                len)
    # Pass through Series() to canonicalize results.
    assert (pd.Series(result) == pd.Series(exp_res)).all()


@pytest.mark.parametrize('serial', trusted_formats)
def test_map_numpy_client_server(iris_server, serial):
    exp_res, result = serialize_query_with_map_builtin_function(iris_server,
                                                                serial,
                                                                np.size)
    # Pass through Series() to canonicalize results.
    assert (pd.Series(result) == pd.Series(exp_res)).all()


@pytest.mark.xfail(reason="pickle does not produce same error")
@pytest.mark.parametrize('serial', all_formats)
def test_builtin_403_exception(iris_server, serial):
    '''
    ensure exception is raised when both map and apply are invoked.
    exception is raised in check_request() when object_hook is invoked;
    this is when the payload is loaded from the bytes object in reqeust.data
    '''
    t = symbol('t', discover(iris))

    for name in ('map', 'apply'):
        func = getattr(t.species, name)
        expr = func(eval, 'int')
        query = {'expr': to_tree(expr)}
        response = iris_server.post('/compute',
                                    data=serial.dumps(query),
                                    headers=mimetype(serial))

        assert '403 FORBIDDEN'.lower() in response.status.lower()


@pytest.mark.xfail(reason="pickle does nto produce same error")
@pytest.mark.parametrize('serial', all_formats)
def test_builtin_501_exception(iris_server, serial):
    t = symbol('t', discover(iris))

    for name in ('map', 'apply'):
        func = getattr(t.species, name)
        expr = func(copy, 'int')
        query = {'expr': to_tree(expr)}
        response = iris_server.post('/compute',
                                    data=serial.dumps(query),
                                    headers=mimetype(serial))

        assert '501 Not Implemented'.lower() in response.status.lower()


@pytest.mark.parametrize('serial', trusted_formats)
def test_map_pandas_client_server(iris_server, serial):
    exp_res, result = serialize_query_with_map_builtin_function(iris_server,
                                                                serial,
                                                                pd.isnull)
    # Pass through Series() to canonicalize results.
    assert (pd.Series(result) == pd.Series(exp_res)).all()



@pytest.mark.parametrize('serial', trusted_formats)
def test_apply_client_server(iris_server, serial):
    test = iris_server
    t = symbol('t', discover(iris))
    expr = t.species.apply(id, 'int') # Very dumb example...
    query = {'expr': to_tree(expr)}
    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))
    assert 'OK' in response.status
    respdata = serial.loads(response.data)
    result = serial.data_loads(respdata['data'])
    assert isinstance(result, _inttypes)
    assert isinstance(compute(expr, {t: iris}), _inttypes)


@pytest.mark.parametrize('serial', all_formats)
def test_server_can_compute_sqlalchemy_reductions(test, serial):
    expr = t.db.iris.petal_length.sum()
    query = {'expr': to_tree(expr)}
    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))

    assert 'OK' in response.status
    respdata = serial.loads(response.data)
    result = serial.data_loads(respdata['data'])
    assert result == into(int, compute(expr, {t: tdata}))
    assert list(respdata['names']) == ['petal_length_sum']


@pytest.mark.parametrize('serial', all_formats)
def test_serialization_endpoints(test, serial):
    expr = t.db.iris.petal_length.sum()
    query = {'expr': to_tree(expr)}
    response = test.post('/compute',
                         data=serial.dumps(query),
                         headers=mimetype(serial))

    assert 'OK' in response.status
    respdata = serial.loads(response.data)
    result = serial.data_loads(respdata['data'])
    assert result == into(int, compute(expr, {t: tdata}))
    assert list(respdata['names']) == ['petal_length_sum']


def test_cors_compute(test):
    res = test.options('/compute')
    assert res.status_code == RC.OK
    assert 'HEAD' in res.headers['Allow']
    assert 'OPTIONS' in res.headers['Allow']
    assert 'POST' in res.headers['Allow']
    # we don't allow gets because we're always sending data
    assert 'GET' not in res.headers['Allow']


def test_cors_datashape(test):
    res = test.options('/datashape')
    assert res.status_code == RC.OK
    assert 'HEAD' in res.headers['Allow']
    assert 'OPTIONS' in res.headers['Allow']
    assert 'GET' in res.headers['Allow']
    # we don't allow posts because we're just getting (meta)data.
    assert 'POST' not in res.headers['Allow']


def test_cors_add(test_add):
    res = test_add.options('/add')
    assert res.status_code == RC.OK
    assert 'HEAD' in res.headers['Allow']
    assert 'POST' in res.headers['Allow']
    assert 'OPTIONS' in res.headers['Allow']
    # we don't allow get because we're sending data.
    assert 'GET' not in res.headers['Allow']


@pytest.fixture(scope='module')
def username():
    return 'blaze-dev'


@pytest.fixture(scope='module')
def password():
    return 'SecretPassword123'


@pytest.fixture(scope='module')
def server_with_auth(username, password):
    def auth(a):
        return a and a.username == username and a.password == password

    s = Server(tdata, all_formats, authorization=auth)
    s.app.testing = True
    return s


@pytest.yield_fixture
def test_with_auth(server_with_auth):
    with server_with_auth.app.test_client() as c:
        yield c


def basic_auth(username, password):
    return b'Basic ' + b64encode(':'.join((username, password)).encode('utf-8'))


@pytest.mark.parametrize('serial', all_formats)
def test_auth(test_with_auth, username, password, serial):
    expr = t.accounts.amount.sum()
    query = {'expr': to_tree(expr)}

    r = test_with_auth.get('/datashape',
                           headers={'authorization': basic_auth(username, password)})
    assert r.status_code == RC.OK
    headers = mimetype(serial)
    headers['authorization'] = basic_auth(username, password)
    s = test_with_auth.post('/compute',
                            data=serial.dumps(query),
                            headers=headers)
    assert s.status_code == RC.OK

    u = test_with_auth.get('/datashape',
                           headers={'authorization': basic_auth(username + 'a', password + 'a')})
    assert u.status_code == RC.UNAUTHORIZED

    headers['authorization'] = basic_auth(username + 'a', password + 'a')
    v = test_with_auth.post('/compute',
                            data=serial.dumps(query),
                            headers=headers)
    assert v.status_code == RC.UNAUTHORIZED


@pytest.mark.parametrize('serial', all_formats)
def test_minute_query(test, serial):
    expr = t.events.when.minute
    query = {'expr': to_tree(expr)}
    result = test.post('/compute',
                       headers=mimetype(serial),
                       data=serial.dumps(query))
    expected = {'data': [0, 0],
                'names': ['when_minute'],
                'datashape': '2 * ?int64'}
    assert result.status_code == RC.OK
    resp = serial.loads(result.data)
    assert list(serial.data_loads(resp['data'])) == expected['data']
    assert list(resp['names']) == expected['names']
    assert resp['datashape'] == expected['datashape']


@pytest.mark.parametrize('serial', all_formats)
def test_isin(test, serial):
    expr = t.events.value.isin(frozenset([1]))
    query = {'expr': to_tree(expr)}
    result = test.post('/compute',
                       headers=mimetype(serial),
                       data=serial.dumps(query))
    expected = {'data': [True, False],
                'names': ['value'],
                'datashape': '2 * bool'}
    assert result.status_code == RC.OK
    resp = serial.loads(result.data)
    assert list(serial.data_loads(resp['data'])) == expected['data']
    assert list(resp['names']) == expected['names']
    assert resp['datashape'] == expected['datashape']


@pytest.mark.parametrize('serial', all_formats)
def test_add_errors(temp_add_server, serial):
    pre_datashape = datashape.dshape(temp_add_server
                                     .get('/datashape')
                                     .data.decode('utf-8'))
    bunk_path = example('bunk.csv')
    blob = serial.dumps({'bunk': bunk_path})
    response1 = temp_add_server.post('/add',
                                     headers=mimetype(serial),
                                     data=blob)
    assert response1.status_code == RC.UNPROCESSABLE_ENTITY

    # Test that the datashape of the server is accessible and unchanged after
    # trying to add a non-existent dataset.
    response2 = temp_add_server.get('/datashape')
    assert response2.status_code == RC.OK
    response_dshape = datashape.dshape(response2.data.decode('utf-8'))
    assert_dshape_equal(pre_datashape, response_dshape)


@pytest.mark.parametrize('serial', all_formats)
def test_add_default_not_allowed(temp_server, serial):
    iris_path = example('iris.csv')
    blob = serial.dumps({'iris': iris_path})
    response1 = temp_server.post('/add',
                                 headers=mimetype(serial),
                                 data=blob)
    assert 'NOT FOUND' in response1.status
    assert response1.status_code == RC.NOT_FOUND


@pytest.mark.parametrize('serial', all_formats)
def test_log_format_exc(temp_server_with_excfmt, serial):
    expr = t.dumb.sort()
    bad_query = {'expr': to_tree(expr)}

    server, log_stream = temp_server_with_excfmt
    result = server.post('/compute', headers=mimetype(serial),
                         data=serial.dumps(bad_query))

    assert 'CUSTOM TRACEBACK' in log_stream.getvalue()



@pytest.mark.parametrize('serial', all_formats)
def test_add_data_to_server(temp_add_server, serial):
    # add data
    iris_path = example('iris.csv')
    blob = serial.dumps({'iris': iris_path})
    response1 = temp_add_server.post('/add',
                                     headers=mimetype(serial),
                                     data=blob)
    assert 'CREATED' in response1.status
    assert response1.status_code == RC.CREATED

    # check for expected server datashape
    response2 = temp_add_server.get('/datashape')
    expected2 = discover({'iris': data(iris_path)})
    response_dshape = datashape.dshape(response2.data.decode('utf-8'))
    assert_dshape_equal(response_dshape.measure.dict['iris'],
                        expected2.measure.dict['iris'])

    # compute on added data
    t = data({'iris': data(iris_path)})
    expr = t.iris.petal_length.sum()

    response3 = temp_add_server.post('/compute',
                                     data=serial.dumps({'expr': to_tree(expr)}),
                                     headers=mimetype(serial))

    result3 = serial.data_loads(serial.loads(response3.data)['data'])
    expected3 = compute(expr, {'iris': data(iris_path)})
    assert result3 == expected3


@pytest.mark.parametrize('serial', all_formats)
def test_cant_add_data_to_server(iris_server, serial):
    # try adding more data to server
    iris_path = example('iris.csv')
    blob = serial.dumps({'iris': iris_path})
    response1 = iris_server.post('/add',
                                 headers=mimetype(serial),
                                 data=blob)
    assert response1.status_code == RC.UNPROCESSABLE_ENTITY


@pytest.mark.parametrize('serial', all_formats)
def test_add_data_twice_error(temp_add_server, serial):
    # add iris
    iris_path = example('iris.csv')
    payload = serial.dumps({'iris': iris_path})
    temp_add_server.post('/add',
                         headers=mimetype(serial),
                         data=payload)

    # Try to add to existing 'iris'
    resp = temp_add_server.post('/add',
                                headers=mimetype(serial),
                                data=payload)
    assert resp.status_code == RC.CONFLICT

    # Verify the server still serves the original 'iris'.
    response_ds = temp_add_server.get('/datashape').data.decode('utf-8')
    ds = datashape.dshape(response_ds)
    t = symbol('t', ds)
    query = {'expr': to_tree(t.iris)}
    resp = temp_add_server.post('/compute',
                                data=serial.dumps(query),
                                headers=mimetype(serial))
    assert resp.status_code == RC.OK


@pytest.mark.parametrize('serial', all_formats)
def test_add_two_data_sets_at_once_error(temp_add_server, serial):
    # Try to add two things at once
    payload = serial.dumps({'foo': 'iris.csv',
                            'bar': 'iris.csv'})
    resp = temp_add_server.post('/add',
                                headers=mimetype(serial),
                                data=payload)
    assert resp.status_code == RC.UNPROCESSABLE_ENTITY


@pytest.mark.parametrize('serial', all_formats)
def test_add_bunk_data_error(temp_add_server, serial):
    # Try to add bunk data
    payload = serial.dumps({'foo': None})
    resp = temp_add_server.post('/add',
                                headers=mimetype(serial),
                                data=payload)
    assert resp.status_code == RC.UNPROCESSABLE_ENTITY


@pytest.mark.parametrize('serial', all_formats)
def test_bad_add_payload(temp_add_server, serial):
    # try adding more data to server
    blob = serial.dumps('This is not a mutable mapping.')
    response1 = temp_add_server.post('/add',
                                     headers=mimetype(serial),
                                     data=blob)
    assert response1.status_code == RC.UNPROCESSABLE_ENTITY


@pytest.mark.parametrize('serial', all_formats)
def test_add_expanded_payload(temp_add_server, serial):
    # Ensure that the expanded payload format is accepted by the server
    iris_path = example('iris.csv')
    blob = serial.dumps({'iris': {'source': iris_path,
                                  'kwargs': {'delimiter': ','}}})
    response1 = temp_add_server.post('/add',
                                     headers=mimetype(serial),
                                     data=blob)
    assert 'CREATED' in response1.status
    assert response1.status_code == RC.CREATED


@pytest.mark.parametrize('serial', all_formats)
def test_add_expanded_payload_with_imports(temp_add_server, serial):
    # Ensure that the expanded payload format is accepted by the server
    iris_path = example('iris.csv')
    blob = serial.dumps({'iris': {'source': iris_path,
                                  'kwargs': {'delimiter': ','},
                                  'imports': ['csv']}})
    response1 = temp_add_server.post('/add',
                                     headers=mimetype(serial),
                                     data=blob)
    assert 'CREATED' in response1.status
    assert response1.status_code == RC.CREATED


@pytest.mark.parametrize('serial', all_formats)
def test_add_expanded_payload_has_effect(temp_add_server, serial):
    # Ensure that the expanded payload format actually passes the arguments
    # through to the resource constructor
    iris_path = example('iris-latin1.tsv')
    csv_kwargs = {'delimiter': '\t', 'encoding': 'iso-8859-1'}
    blob = serial.dumps({'iris': {'source': iris_path,
                                  'kwargs': csv_kwargs}})
    response1 = temp_add_server.post('/add',
                                     headers=mimetype(serial),
                                     data=blob)
    assert 'CREATED' in response1.status
    assert response1.status_code == RC.CREATED

    # check for expected server datashape
    response2 = temp_add_server.get('/datashape')
    expected2 = discover({'iris': data(iris_path, **csv_kwargs)})
    response_dshape = datashape.dshape(response2.data.decode('utf-8'))
    assert_dshape_equal(response_dshape.measure.dict['iris'],
                        expected2.measure.dict['iris'])

    # compute on added data
    t = data({'iris': data(iris_path, **csv_kwargs)})
    expr = t.iris.petal_length.sum()

    response3 = temp_add_server.post('/compute',
                                     data=serial.dumps({'expr': to_tree(expr)}),
                                     headers=mimetype(serial))

    result3 = serial.data_loads(serial.loads(response3.data)['data'])
    expected3 = compute(expr, {'iris': data(iris_path, **csv_kwargs)})
    assert result3 == expected3


@pytest.mark.parametrize('serial', all_formats)
def test_odo_kwargs(test, serial):
    expr = t.dumb
    bad_query = {'expr': to_tree(expr)}

    result = test.post('/compute',
                       headers=mimetype(serial),
                       data=serial.dumps(bad_query))
    assert result.status_code == RC.INTERNAL_SERVER_ERROR
    assert b'return_df must be passed' in result.data

    good_query = {'expr': to_tree(expr),
                  'odo_kwargs': {'return_df': odo(DumbResource.df, list)}}
    result = test.post('/compute',
                       headers=mimetype(serial),
                       data=serial.dumps(good_query))
    assert result.status_code == RC.OK
    tdata = serial.loads(result.data)
    dshape = discover(DumbResource.df)
    assert_dshape_equal(datashape.dshape(tdata['datashape']),
                        dshape)
    assert_frame_equal(odo(serial.data_loads(tdata['data']),
                           DataFrame,
                           dshape=dshape),
                       DumbResource.df)


@pytest.mark.parametrize('serial', all_formats)
def test_compute_kwargs(test, serial):
    expr = t.dumb.sort()
    bad_query = {'expr': to_tree(expr)}

    result = test.post('/compute',
                       headers=mimetype(serial),
                       data=serial.dumps(bad_query))
    assert result.status_code == RC.INTERNAL_SERVER_ERROR
    assert b'return_df must be passed' in result.data

    good_query = {'expr': to_tree(expr),
                  'compute_kwargs': {'return_df': odo(DumbResource.df, list)}}
    result = test.post('/compute',
                       headers=mimetype(serial),
                       data=serial.dumps(good_query))
    assert result.status_code == RC.OK
    tdata = serial.loads(result.data)
    dshape = discover(DumbResource.df)
    assert_dshape_equal(datashape.dshape(tdata['datashape']),
                        dshape)
    assert_frame_equal(odo(serial.data_loads(tdata['data']),
                           DataFrame,
                           dshape=dshape),
                       DumbResource.df)


def test_fastmsgmpack_mutable_dataframe(test):
    expr = t.events  # just get back the dataframe
    query = {'expr': to_tree(expr)}
    result = test.post('/compute',
                       headers=mimetype(fastmsgpack),
                       data=fastmsgpack.dumps(query))
    assert result.status_code == RC.OK
    data = fastmsgpack.data_loads(fastmsgpack.loads(result.data)['data'])

    for block in data._data.blocks:
        # make sure all the blocks are mutable
        assert block.values.flags.writeable


@pytest.mark.parametrize('serial', all_formats)
def test_bad_payload_keys(test, serial):
    result = test.post('/compute',
                       headers=mimetype(serial),
                       data=serial.dumps({u'ayy': 'lmao',
                                          u'nah': 'fam',
                                          u'profile': True,
                                          u'profiler_output': ':response'}))

    assert result.status_code == RC.BAD_REQUEST
    assert (result.data ==
            "unexpected keys in payload: [{u}'ayy', {u}'nah']".format(
                u='u' if PY2 else '',
            ).encode('ascii'))



from odo import chunks
from blaze import discover, into, compute, symbol
from datashape.predicates import iscollection


L = [1, 2, 3, 4, 5, 6]
cL = chunks(list)([[1., 2., 3.], [4., 5., 6.]])
s = symbol('s', discover(cL))


def test_chunks_compute():
    exprs = [s, s + 1, s.max(), s.mean() + 1, s.head()]
    for e in exprs:
        result = compute(e, {s: cL})
        expected = compute(e, {s: L})
        if iscollection(e.dshape):
            result = into(list, result)
            expected = into(list, expected)
        assert result == expected


def test_chunks_head():
    assert compute(s.head(2), cL) == (1., 2.)


def test_pmap_default():
    flag = [0]

    def mymap(func, seq):
        flag[0] = True
        return map(func, seq)

    from blaze import set_default_pmap
    set_default_pmap(mymap)

    compute(s + 1, cL)

    assert flag[0] is True



from __future__ import absolute_import, division, print_function

import datetime
from collections import Iterator
from itertools import islice, product
import os
import re

try:
    from cytoolz import nth, unique, concat
except ImportError:
    from toolz import nth, unique, concat

import numpy as np
# these are used throughout blaze, don't remove them
from odo.utils import tmpfile, filetext, filetexts, raises, keywords, ignoring
import pandas as pd
import psutil
import sqlalchemy as sa

# Imports that replace older utils.
from .compatibility import map, zip


def nth_list(n, seq):
    """

    >>> tuple(nth_list([0, 1, 4], 'Hello'))
    ('H', 'e', 'o')
    >>> tuple(nth_list([4, 1, 0], 'Hello'))
    ('o', 'e', 'H')
    >>> tuple(nth_list([0, 0, 0], 'Hello'))
    ('H', 'H', 'H')
    """
    seq = iter(seq)

    result = []
    old = 0
    item = next(seq)
    for index in sorted(n):
        for i in range(index - old):
            item = next(seq)
        result.append(item)
        old = index

    order = [x[1] for x in sorted(zip(n, range(len(n))))]
    return (result[i] for i in order)


def get(ind, coll, lazy=False):
    """

    >>> get(0, 'Hello')
    'H'

    >>> get([1, 0], 'Hello')
    ('e', 'H')

    >>> get(slice(1, 4), 'Hello')
    ('e', 'l', 'l')

    >>> get(slice(1, 4), 'Hello', lazy=True)
    <itertools.islice object at ...>
    """
    if isinstance(ind, list):
        result = nth_list(ind, coll)
    elif isinstance(ind, slice):
        result = islice(coll, ind.start, ind.stop, ind.step)
    else:
        if isinstance(coll, Iterator):
            result = nth(ind, coll)
        else:
            result = coll[ind]
    if not lazy and isinstance(result, Iterator):
        result = tuple(result)
    return result


def ndget(ind, data):
    """
    Get from N-Dimensional getable

    Can index with elements, lists, or slices.  Mimic's numpy fancy indexing on
    generic indexibles.

    >>> data = [[[1, 2], [3, 4]], [[5, 6], [7, 8]]]
    >>> ndget(0, data)
    [[1, 2], [3, 4]]
    >>> ndget((0, 1), data)
    [3, 4]
    >>> ndget((0, 0, 0), data)
    1
    >>> ndget((slice(0, 2), [0, 1], 0), data)
    ((1, 3), (5, 7))
    """
    if isinstance(ind, tuple) and len(ind) == 1:
        ind = ind[0]
    if not isinstance(ind, tuple):
        return get(ind, data)
    result = get(ind[0], data)
    if isinstance(ind[0], (list, slice)):
        return type(result)(ndget(ind[1:], row) for row in result)
    else:
        return ndget(ind[1:], result)


def normalize_to_date(dt):
    if isinstance(dt, datetime.datetime) and not dt.time():
        return dt.date()
    else:
        return dt


def assert_allclose(lhs, rhs):
    for tb in map(zip, lhs, rhs):
        for left, right in tb:
            if isinstance(left, (np.floating, float)):
                # account for nans
                assert np.all(np.isclose(left, right, equal_nan=True))
                continue
            if isinstance(left, datetime.datetime):
                left = normalize_to_date(left)
            if isinstance(right, datetime.datetime):
                right = normalize_to_date(right)
            assert left == right


def example(filename, datapath=os.path.join('examples', 'data')):
    import blaze
    return os.path.join(os.path.dirname(blaze.__file__), datapath, filename)


def available_memory():
    return psutil.virtual_memory().available


def listpack(x):
    """
    >>> listpack(1)
    [1]
    >>> listpack((1, 2))
    [1, 2]
    >>> listpack([1, 2])
    [1, 2]
    """
    if isinstance(x, tuple):
        return list(x)
    elif isinstance(x, list):
        return x
    else:
        return [x]


def normalize(s):
    """Normalize a sql expression for comparison in tests.

    Parameters
    ----------
    s : str or Selectable
        The expression to normalize. If this is a selectable, it will be
        compiled with literals inlined.

    Returns
    -------
    cs : Any
        An object that can be compared against another normalized sql
        expression.
    """
    if isinstance(s, sa.sql.Selectable):
        s = literal_compile(s)
    s = re.sub(r'(\(|\))', r' \1 ', s)       # normalize spaces around parens
    s = ' '.join(s.strip().split()).lower()  # normalize whitespace and case
    s = re.sub(r'(alias)_?\d*', r'\1', s)    # normalize aliases
    return re.sub(r'__([A-Za-z_][A-Za-z_0-9]*)', r'\1', s)


def literal_compile(s):
    """Compile a sql expression with bind params inlined as literals.

    Parameters
    ----------
    s : Selectable
        The expression to compile.

    Returns
    -------
    cs : str
        An equivalent sql string.
    """
    return str(s.compile(compile_kwargs={'literal_binds': True}))


def ordered_intersect(*sets):
    """Set intersection of two sequences that preserves order.

    Parameters
    ----------
    sets : tuple of Sequence

    Returns
    -------
    generator

    Examples
    --------
    >>> list(ordered_intersect('abcd', 'cdef'))
    ['c', 'd']
    >>> list(ordered_intersect('bcda', 'bdfga'))
    ['b', 'd', 'a']
    >>> list(ordered_intersect('zega', 'age'))  # 1st sequence determines order
    ['e', 'g', 'a']
    >>> list(ordered_intersect('gah', 'bag', 'carge'))
    ['g', 'a']
    """
    common = frozenset.intersection(*map(frozenset, sets))
    return (x for x in unique(concat(sets)) if x in common)


class attribute(object):
    """An attribute that can be overridden by instances.
    This is like a non data descriptor property.

    Parameters
    ----------
    f : callable
        The function to execute.
    """
    def __init__(self, f):
        self._f = f

    def __get__(self, instance, owner):
        if instance is None:
            return self

        return self._f(instance)


def parameter_space(*args):
    """Unpack a sequence of positional parameter spaces into the product of each
    space.

    Parameters
    ----------
    *args
        The parameters spaces to create a product of.

    Returns
    -------
    param_space : tuple[tuple]
        The product of each of the spaces.

    Examples
    --------
    # trivial case
    >>> parameter_space(0, 1, 2)
    ((0, 1, 2),)

    # two 2-tuples
    >>> parameter_space((0, 1), (2, 3))
    ((0, 2), (0, 3), (1, 2), (1, 3))

    Notes
    -----
    This is a convenience for passing to :func:`pytest.mark.parameterized`
    """
    return tuple(product(*(
        arg if isinstance(arg, tuple) else (arg,) for arg in args
    )))
